module allocate_module
  ! -- modules
  use kind_module, only: DP, I4
  use constvalue_module, only: DZERO, DONE
  use read_inputfile, only: maxout_iter, solver_flag
  use set_classification, only: ncalc, ncals, nunknow
  use set_condition, only: off_index
  use set_calculation, only: cell_ksx, cell_ksy, cell_ksz

  implicit none
  private
  public :: allocate_varts
  real(DP), allocatable, public :: surf_old(:), rech_old(:), well_old(:)
  real(DP), allocatable, public :: calc_rech(:), calc_well(:)

  real(DP), allocatable, public :: head_old(:), head_new(:), head_pre(:)
  real(DP), allocatable, public :: dmat(:), lumat(:), rhs(:)
  real(DP), allocatable, public :: srati_old(:), srati_new(:), pertur_srati(:)
  real(DP), allocatable, public :: hydcon_new(:,:), pertur_hyd(:,:)
  real(DP), allocatable, public :: rel_perm(:), pertur_rel(:)
  real(DP), allocatable, public :: head_change(:), grad(:)

  real(DP), allocatable, public :: x(:), resi(:), e(:)

  ! -- local
  integer(I4) :: nlumat, i

  contains

  subroutine allocate_varts()
  !*****************************************************************************
  ! allocate_varts -- Allocate variable for time step
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------

    nlumat = off_index(ncalc)

    ! -- Allocate timeupdate (timeup)
      call allocate_timeup()

    ! -- Allocate matrix & vector (matvec)
      call allocate_matvec()

    ! -- Allocate linear algebra (linalge)
      call allocate_linalge()

    return

  end subroutine allocate_varts

  subroutine allocate_timeup()
  !*****************************************************************************
  ! allocate_timeup -- Allocate timeupdate
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(surf_old(ncals), rech_old(ncals), well_old(ncalc))
    surf_old(:) = DZERO ; rech_old(:) = DZERO ; well_old(:) = DZERO
    allocate(calc_rech(ncals), calc_well(ncalc))
    calc_rech(:) = DZERO ; calc_well(:) = DZERO

    return

  end subroutine allocate_timeup

  subroutine allocate_matvec()
  !*****************************************************************************
  ! allocate_matvec -- Allocate matrix & vector
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(head_old(ncalc), head_new(ncalc), head_pre(ncalc))
    head_old(:) = DZERO ; head_new(:) = DZERO ; head_pre(:) = DZERO
    allocate(dmat(ncalc), lumat(nlumat), rhs(ncalc))
    dmat(:) = DZERO ; lumat(:) = DZERO ; rhs(:) = DZERO
    allocate(srati_old(ncalc), srati_new(ncalc), pertur_srati(ncalc))
    srati_old(:) = DZERO ; srati_new(:) = DZERO ; pertur_srati(:) = DZERO
    allocate(hydcon_new(ncalc,3), pertur_hyd(ncalc,3))
    hydcon_new(:,:) = DZERO ; pertur_hyd(:,:) = DZERO
    allocate(rel_perm(ncalc), pertur_rel(ncalc))
    rel_perm(:) = DZERO ; pertur_rel(:) = DZERO
    allocate(head_change(nunknow), grad(nunknow))
    head_change(:) = DZERO ; grad(:) = DZERO

    do i = 1, ncalc
      hydcon_new(i,1) = cell_ksz(i)
      hydcon_new(i,2) = cell_ksy(i)
      hydcon_new(i,3) = cell_ksx(i)
    end do

    deallocate(cell_ksx, cell_ksy, cell_ksz)

    return

  end subroutine allocate_matvec

  subroutine allocate_linalge()
  !*****************************************************************************
  ! allocate_linalge -- Allocate linear algebra
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(resi(ncalc))
    resi(:) = DZERO

    return

  end subroutine allocate_linalge

end module allocate_module
