module make_output
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE, VARIABLELENGTH, OUTPUTLENGTH, FACE,&
                               DHALF
  use read_inputfile, only: nx, ny, nz, totaln_wlout, wlout_i, wlout_j, wlout_k,&
                            wlout_flag, totaln_msout, msout_class_name,&
                            msout_i, msout_j, msout_k, msout_flag, sim_flag
  use set_classification, only: calc2cell, cell2calc, calc2cals, cals2cell,&
                                ncalc, nunknow, cals_end
  use make_cell, only: area, dis
  use set_condition, only: off_index, off_row, conn_dir, set_point2out,&
                           set_class2out
  use set_calculation, only: bound_cflag, rech_cflag, well_cflag, surf_cflag,&
                             cell_ss, cell_sat, surf_head, ghost_head,&
                             cell_well, cell_rech
  use calc_parameter, only: calc_relperm, calc_srati, calc_hyd_geoharm,&
                            calc_hyd_upwind, calc_hyd_harm, calc_hyd_geo,&
                            calc_hyd_arith
  use allocate_module, only: head_new, head_old, srati_new, srati_old,&
                             hydcon_new, calc_well, calc_rech, rel_perm
  use time_module, only: delt
  use calc_massbalance, only: calc_mass

  implicit none
  private
  public :: set_pointwlhead, set_classmshead, set_pointmshead, set_wholemshead
  public :: allocate_wlevel, allocate_mass, allocate_velc, allocate_welldif
  public :: calc_pointwl, calc_cellmass, calc_outmass, calc_outvelc
  public :: calc_welldif, calc_rechdif, allocate_rechdif
  integer(I4), allocatable, public :: well_diflag(:), rech_diflag(:)
  character(len=OUTPUTLENGTH), public :: wl_header, ms_header
  real(DP), allocatable, public :: wlevel(:), rec_sum(:), wel_sum(:)
  real(DP), allocatable, public :: sto_sum(:), con_sum(:), all_sum(:)
  real(DP), allocatable, public :: vel_point(:,:)
  real(DP), allocatable, public :: inp_we(:), act_we(:), dif_we(:)
  real(DP), allocatable, public :: inp_re(:), act_re(:), dif_re(:)
  integer(I4), public :: wlout_tnum, msout_tnum, we_dnum, re_dnum

  ! -- local
  integer(I4) :: i, j, k, c, s
  integer(I4), allocatable :: wlout_wri(:), wlout_wrj(:), wlout_wrk(:)
  integer(I4), allocatable :: point_num(:), wlout_num(:), msout_num(:)
  integer(I4), allocatable :: ms_index(:), ms_cnum(:)
  real(DP), allocatable :: rech_ms(:), well_ms(:), stor_ms(:), conn_ms(:)
  real(DP), allocatable :: vel_face(:,:)

  contains

  subroutine set_pointwlhead()
  !*****************************************************************************
  ! set_pointwlhead -- Set point water level header
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    character(len=VARIABLELENGTH) :: str_wli, str_wlj, str_wlk
    !---------------------------------------------------------------------------
    ! allocate
    allocate(wlout_wri(nunknow), wlout_wrj(nunknow), wlout_wrk(nunknow))
    wlout_wri(:) = 0 ; wlout_wrj(:) = 0 ; wlout_wrk(:) = 0
    allocate(point_num(nunknow))
    point_num(:) = 0

    wlout_tnum = 0

    call set_point2out(totaln_wlout, wlout_i, wlout_j, wlout_k, wlout_wri, wlout_wrj, wlout_wrk, point_num, wlout_tnum)

    allocate(wlout_num(wlout_tnum))
    wlout_num(:) = 0

    do i = 1, wlout_tnum
      write(str_wli, *) wlout_wri(i)
      write(str_wlj, *) wlout_wrj(i)
      if (wlout_flag == 1) then
        write(str_wlk, *) wlout_wrk(i)
      else if (wlout_flag == 2) then
        write(str_wlk, *) -1
      end if
      if (i == 1) then
        wl_header = trim(adjustl(str_wli)) // "_" // trim(adjustl(str_wlj))&
                    // "_" // trim(adjustl(str_wlk))
      else
        wl_header = trim(adjustl(wl_header)) // "," // trim(adjustl(str_wli))&
                    // "_" // trim(adjustl(str_wlj)) // "_" // trim(adjustl(str_wlk))
      end if
      wlout_num(i) = point_num(i)
    end do

    deallocate(wlout_wrk, point_num)

    if (wlout_flag /= 2) then
      deallocate(wlout_wri, wlout_wrj)
    end if

    return

  end subroutine set_pointwlhead

  subroutine allocate_wlevel()
  !*****************************************************************************
  ! allocate_wlevel -- Allocate waterlevel
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    allocate(wlevel(wlout_tnum))
    wlevel = DZERO

    return

  end subroutine allocate_wlevel

  subroutine calc_pointwl()
  !*****************************************************************************
  ! calc_pointwl -- Calculate point water level
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: c_num
    !---------------------------------------------------------------------------
    do i = 1, wlout_tnum
      if (wlout_flag == 1) then
        c = wlout_num(i)
        c_num = calc2cell(c)
        wlevel(i) = head_new(c)
      else
        do k = nz, 1, -1
          c_num = nx*ny*(k-1) + nx*(wlout_wrj(i)-1) + wlout_wri(i)
          c = cell2calc(c_num)
          if (srati_new(c) < DONE) then
            exit
          end if
        end do
        wlevel(i) = head_new(c)
      end if
    end do

    return

  end subroutine calc_pointwl

  subroutine set_classmshead()
  !*****************************************************************************
  ! set_classmshead -- Set classification massbalance header
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: ms_tnum
    !---------------------------------------------------------------------------
    allocate(ms_index(0:totaln_msout))
    ms_index(:) = 0
    allocate(ms_cnum(totaln_msout*nunknow))
    ms_cnum(:) = 0

    ms_tnum = 0

    call set_class2out(totaln_msout, msout_class_name, ms_index, ms_cnum, ms_tnum)

    allocate(msout_num(ms_tnum))
    msout_num(:) = 0

    do i = 1, totaln_msout
      if (i == 1) then
        ms_header = trim(adjustl(msout_class_name(i)))
      else
        ms_header = trim(adjustl(ms_header)) // ",,,,,"// trim(adjustl(msout_class_name(i)))
      end if
    end do

    do i = 1, ms_tnum
      msout_num(i) = ms_cnum(i)
    end do

    msout_tnum = totaln_msout

    deallocate(ms_cnum)

    return

  end subroutine set_classmshead

  subroutine set_pointmshead()
  !*****************************************************************************
  ! set_pointmshead -- Set point massbalance header
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4), allocatable :: msout_wri(:), msout_wrj(:), msout_wrk(:)
    character(len=VARIABLELENGTH) :: str_msi, str_msj, str_msk
    !---------------------------------------------------------------------------
    allocate(msout_wri(nunknow), msout_wrj(nunknow), msout_wrk(nunknow))
    msout_wri(:) = 0 ; msout_wrj(:) = 0 ; msout_wrk(:) = 0
    allocate(point_num(nunknow))
    point_num(:) = 0

    msout_tnum = 0

    call set_point2out(totaln_msout, msout_i, msout_j, msout_k, msout_wri, msout_wrj, msout_wrk, point_num, msout_tnum)

    allocate(msout_num(msout_tnum))
    msout_num(:) = 0

    do i = 1, msout_tnum
      write(str_msi, *) msout_wri(i)
      write(str_msj, *) msout_wrj(i)
      write(str_msk, *) msout_wrk(i)
      if (i == 1) then
        ms_header = trim(adjustl(str_msi)) // "_" // trim(adjustl(str_msj))&
                  // "_" // trim(adjustl(str_msk))
      else
        ms_header = trim(adjustl(ms_header)) // ",,,,,"// trim(adjustl(str_msi))&
                  // "_" // trim(adjustl(str_msj)) // "_" // trim(adjustl(str_msk))
      end if
      msout_num(i) = point_num(i)
    end do

    deallocate(msout_wri, msout_wrj, msout_wrk, point_num)

    return

  end subroutine set_pointmshead

  subroutine set_wholemshead()
  !*****************************************************************************
  ! set_wholemshead -- Set whole area massbalance header
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    msout_tnum = 1

    allocate(msout_num(msout_tnum))
    msout_num(:) = 0

    msout_num(1) = 1

    ms_header = "whole area"

    return

  end subroutine set_wholemshead

  subroutine allocate_mass()
  !*****************************************************************************
  ! allocate_mass -- Allocate massbalance
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    allocate(rech_ms(cals_end(1)), well_ms(nunknow), stor_ms(nunknow), conn_ms(nunknow))
    rech_ms(:) = DZERO ; well_ms(:) = DZERO ; stor_ms(:) = DZERO ; conn_ms(:) = DZERO
    allocate(rec_sum(msout_tnum), wel_sum(msout_tnum), sto_sum(msout_tnum))
    allocate(con_sum(msout_tnum), all_sum(msout_tnum))
    rec_sum(:) = DZERO ; wel_sum(:) = DZERO ; sto_sum(:) = DZERO
    con_sum(:) = DZERO ; all_sum(:) = DZERO

    return

  end subroutine allocate_mass

  subroutine calc_cellmass()
  !*****************************************************************************
  ! calc_cellmass -- Calculate cell massbalance
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calculate massbalance (mass)
      call calc_mass(0, nunknow, head_new, head_old, rech_ms, well_ms, stor_ms, conn_ms)

    return

  end subroutine calc_cellmass

  subroutine calc_outmass()
  !*****************************************************************************
  ! calc_outmass -- Calculate output massbalance
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (msout_flag == 0) then
      do i = 1, totaln_msout
        rec_sum(i) = DZERO ; wel_sum(i) = DZERO ; sto_sum(i) = DZERO
        con_sum(i) = DZERO ; all_sum(i) = DZERO
        do k = ms_index(i-1)+1, ms_index(i)
          j = msout_num(k)
          s = calc2cals(j)
          if (s /= 0) then
            rec_sum(i) = rec_sum(i) + rech_ms(s)
          end if
          wel_sum(i) = wel_sum(i) + well_ms(j)
          sto_sum(i) = sto_sum(i) + stor_ms(j)
          con_sum(i) = con_sum(i) + conn_ms(j)
        end do
        all_sum(i) = rec_sum(i) + wel_sum(i) + sto_sum(i) + con_sum(i)
      end do

    else if (msout_flag == 1) then
      do i = 1, msout_tnum
        rec_sum(i) = DZERO ; wel_sum(i) = DZERO ; sto_sum(i) = DZERO
        con_sum(i) = DZERO ; all_sum(i) = DZERO
        j = msout_num(i)
        s = calc2cals(j)
        if (s /= 0) then
          rec_sum(i) = rec_sum(i) + rech_ms(s)
        end if
        wel_sum(i) = wel_sum(i) + well_ms(j)
        sto_sum(i) = sto_sum(i) + stor_ms(j)
        con_sum(i) = con_sum(i) + conn_ms(j)
        all_sum(i) = rec_sum(i) + wel_sum(i) + sto_sum(i) + con_sum(i)
      end do

    else if (msout_flag == 2) then
      rec_sum(1) = DZERO ; wel_sum(1) = DZERO ; sto_sum(1) = DZERO
      con_sum(1) = DZERO ; all_sum(1) = DZERO
      do i = 1, nunknow
        if (bound_cflag(i) /= 1) then
          s = calc2cals(i)
          if (s /= 0) then
            rec_sum(1) = rec_sum(1) + rech_ms(s)
          end if
          wel_sum(1) = wel_sum(1) + well_ms(i)
          sto_sum(1) = sto_sum(1) + stor_ms(i)
          con_sum(1) = con_sum(1) + conn_ms(i)
        end if
      end do
      all_sum(1) = rec_sum(1) + wel_sum(1) + sto_sum(1) + con_sum(1)
    end if

    do i = 1, nunknow
      well_ms(i) = DZERO ; stor_ms(i) = DZERO ; conn_ms(i) = DZERO
      s = calc2cals(i)
      if (s /= 0) then
        rech_ms(s) = DZERO
      end if
    end do

    return

  end subroutine calc_outmass

  subroutine allocate_velc()
  !*****************************************************************************
  ! allocate_velc -- Allocate velocity
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    allocate(vel_point(nunknow,3), vel_face(nunknow,FACE))
    vel_point(:,:) = DZERO ; vel_face(:,:) = DZERO

    return

  end subroutine allocate_velc

  subroutine calc_outvelc()
  !***************************************************************************
  ! calc_outvelc -- Calculate output velocity
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: s, c, dir, d
    real(DP) :: delhead, dis1, dis2, dis12, conn_hyd
    real(DP) :: hyd_face, hyd_face1, hyd_face2, relat
    !---------------------------------------------------------------------------
    ! -- Calculate relative permeability (relperm)
      call calc_relperm(ncalc, DZERO, head_new, srati_new, rel_perm)

    do i = 1, nunknow
      if (bound_cflag(i) /= 1) then
        s = calc2cals(i)
        c = calc2cell(i)
        if (s /= 0) then
          dir = 1
          delhead = head_new(i) - surf_head(s)
          vel_face(i,4-dir) = hydcon_new(i,dir)*delhead/dis(c,dir)*rel_perm(i)
        end if
        do k = off_index(i-1)+1, off_index(i)
          dir = conn_dir(k)
          if (dir <= 3) then
            j = off_row(k)
            dis1 = dis(c,dir)
            delhead = head_new(j) - head_new(i)

            if (bound_cflag(j) == 1) then
              dis2 = dis1
              dis12 = dis1
              j = i
            else
              dis2 = dis(calc2cell(j),(7-dir))
              dis12 = dis1 + dis2
            end if

            ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
              call calc_hyd_upwind(head_new(i), head_new(j), rel_perm(i), rel_perm(j), relat)

            hyd_face1 = hydcon_new(i,dir)
            hyd_face2 = hydcon_new(j,dir)

            ! -- Calculate hydradulic conductivity by harmonic mean (hyd_harm)
              call calc_hyd_harm(hyd_face1, hyd_face2, dis1, dis2, hyd_face)

            conn_hyd = hyd_face*relat

            vel_face(c,4-dir) = conn_hyd*delhead/dis12

          end if
        end do

        do d = 1, 3
          if (vel_face(i,d) > DZERO .and. vel_face(i,7-d) < DZERO) then
            vel_point(i,4-d) = (vel_face(i,d) - vel_face(i,7-d))*DHALF
          else if (vel_face(i,d) < DZERO .and. vel_face(i,7-d) > DZERO) then
            vel_point(i,4-d) = (vel_face(i,d) - vel_face(i,7-d))*DHALF
          else
            vel_point(i,4-d) = vel_face(i,d) - vel_face(i,7-d)
          end if
          if (d == 1) then
            vel_point(i,4-d) = -vel_point(i,4-d)
          end if
        end do
      end if
    end do

    return

  end subroutine calc_outvelc

  subroutine allocate_welldif()
  !*****************************************************************************
  ! allocate_welldif -- Allocate well difference
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    allocate(inp_we(nunknow), act_we(nunknow), dif_we(nunknow))
    inp_we(:) = DZERO ; act_we(:) = DZERO ; dif_we(:) = DZERO
    allocate(well_diflag(nunknow))
    well_diflag(:) = 0

    return

  end subroutine allocate_welldif

  subroutine calc_welldif()
  !***************************************************************************
  ! calc_welldif -- Calculate well difference
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    do i = 1, nunknow
      if (well_cflag(i) == 1) then
        inp_we(i) = inp_we(i) + cell_well(i)*delt
        act_we(i) = act_we(i) + calc_well(i)*delt
        if (inp_we(i) /= act_we(i) .and. well_diflag(i) == 0) then
          we_dnum = we_dnum + 1
          well_diflag(i) = 1
          dif_we(i) = act_we(i) - inp_we(i)
        else if (inp_we(i) /= act_we(i)) then
          dif_we(i) = act_we(i) - inp_we(i)
        end if
      end if
    end do

    return

  end subroutine calc_welldif

  subroutine allocate_rechdif()
  !*****************************************************************************
  ! allocate_rechdif -- Allocate recharge difference
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    allocate(inp_re(cals_end(1)), act_re(cals_end(1)), dif_re(cals_end(1)))
    inp_re(:) = DZERO ; act_re(:) = DZERO ; dif_re(:) = DZERO
    allocate(rech_diflag(cals_end(1)))
    rech_diflag(:) = 0

    return

  end subroutine allocate_rechdif

  subroutine calc_rechdif()
  !***************************************************************************
  ! calc_rechdif -- Calculate recharge difference
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    do i = 1, cals_end(1)
      c = cals2cell(i)
      if (rech_cflag(i) == 1 .and. surf_cflag(i) /= 1 .and. bound_cflag(i) /= 1) then
        inp_re(i) = inp_re(i) + cell_rech(i)*delt*area(c,1)
        act_re(i) = act_re(i) + calc_rech(i)*delt*area(c,1)
        if (inp_re(i) /= act_re(i) .and. rech_diflag(i) == 0) then
          re_dnum = re_dnum + 1
          rech_diflag(i) = 1
          dif_re(i) = act_re(i) - inp_re(i)
        else if (inp_re(i) /= act_re(i)) then
          dif_re(i) = act_re(i) - inp_re(i)
        end if
      end if
    end do

    return

  end subroutine calc_rechdif

end module make_output
