program msg
! ******************************************************************************
! Main program of Multi-scale&time groundwater flow model (Msg)
! ******************************************************************************
! ------------------------------------------------------------------------------
  ! -- modules
  use kind_module, only: I4, DP
  use read_inputfile, only: read_mainf
  use set_calculation, only: prepare_calc
  use allocate_module, only: allocate_varts
  use time_module, only: update_tstep, conv_flag
  use nonlinear_solution, only: calc_numsol
  use write_outputfile, only: write_outf
  use check_simulation, only: check_lastts, lasttime_flag

  implicit none
  ! -- local
  integer(I4) :: i
  integer(I4) :: sta_value(8), end_value(8)
  real(DP) :: calc_stime, calc_etime
  ! -- format
  10 format(/"Run start date and time(yyyy/mm/dd hh:mm:ss) : ",&
            i4,"/",i2.2,"/",i2.2,1x,i2,":",i2.2,":",i2.2,/)
  11 format(/"Run end date and time(yyyy/mm/dd hh:mm:ss) : ",&
            i4,"/",i2.2,"/",i2.2,1x,i2,":",i2.2,":",i2.2,/)
  12 format(/"Calaculation cpu time : ", es12.6, " (sec)")
  ! ----------------------------------------------------------------------------
  ! -- Start time
    call DATE_AND_TIME(values = sta_value)
    write(*,10) (sta_value(i), i = 1, 3), (sta_value(i), i = 5, 7)

  ! -- Read main file (mainf)
    call read_mainf()

  ! -- Prepare calculation (calc)
    call prepare_calc()

  ! -- Allocate variable for time step (varts)
    call allocate_varts()

  ! -- Calculation start time
    call CPU_TIME(calc_stime)

  ! start time step loop
  tstep_loop: do

    ! -- Update time step (tstep)
      call update_tstep()

    ! -- Calculate numerical solution (numsol)
      call calc_numsol()

    if (conv_flag == 1) then
      ! -- Check last time step conditions (lastts)
        call check_lastts()
      if(lasttime_flag == 1) then
        ! -- Write output file (outf)
          call write_outf()
        exit tstep_loop
      else
        ! -- Write output file (outf)
          call write_outf()
      end if
    end if

  end do tstep_loop

  ! -- Calculation start time
    call CPU_TIME(calc_etime)

  write(*,12) calc_etime - calc_stime

  ! -- End time
    call DATE_AND_TIME(values = end_value)
    write(*,11) (end_value(i), i = 1, 3), (end_value(i), i = 5, 7)

end program msg
