module read_inputfile
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: FILELENGTH, VARIABLELENGTH, CHARACTERNUMBER,&
                               TIMELENGTH, DZERO, DONE, DINFI, MINUTESEC,&
                               HOURSEC, DAYSEC, YEARSEC, PARANUMBER

  implicit none
  private
  public :: read_mainf
  !main file
  integer(I4), public :: sim_flag = -1 !simluation flag
  integer(I4), public :: start_flag = -1 !start time flag
  integer(I4), public :: calcout_flag = -1 !calculation flag
  character(len=VARIABLELENGTH), public :: calcout_name
  integer(I4), public :: nx = 0, ny = 0, nz = 0, nxyz = 0
  real(DP), public :: init_step = DZERO, max_step = DZERO, incr_multi = DZERO
  real(DP), public :: end_time = DZERO, calc_multi = DZERO
  character(len=TIMELENGTH), public :: calc_unit
  !type :: inflagtype
    integer(I4), public :: para_flag = -1 !parameter flag
    integer(I4), public :: init_flag = -1 !initial flag
    integer(I4), public :: rech_flag = -1 !recharge flag
    integer(I4), public :: well_flag = -1 !well flag
    integer(I4), public :: bound_flag = -1 !boundary flag
    integer(I4), public :: surf_flag = -1 !surface water flag
    integer(I4), public :: seep_flag = -1 !seepage setting flag
    integer(I4), public :: wlout_flag = -1 !waterlevel output flag
    integer(I4), public :: msout_flag = -1 !massbalance output flag
  !end type inflagtype

  !input list file
  integer(I4), public :: maxout_iter = 0, maxinn_iter = 0, solver_flag = 0
  integer(I4), public :: nlevel = 0, maxvcy_iter = 0, temp_maxinn_iter = 0
  integer(I4), public :: outche_flag = 0, max_swp = 0
  real(DP), public :: criteria = DZERO, switch_criteria = DZERO
  real(DP), public :: newper = DZERO, newper_inv = DZERO
  real(DP), public :: errtol = DZERO, sor_omega = DZERO, sip_alpha = DZERO
  !type :: gridtype
    real(DP), allocatable, public :: xpoint(:,:), ypoint(:,:)
    real(DP), allocatable, public :: zpoint(:,:,:)
  !end type gridtype
  !type :: classtype
    integer(I4), public :: totaln_class = 0
    character(len=VARIABLELENGTH), public :: class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: class_num(:)
    integer(I4), allocatable, public :: class_i(:,:), class_j(:,:), class_k(:,:)
  !end type classtype
  !type :: retentype
    integer(I4), public :: totaln_reten = 0
    character(len=VARIABLELENGTH), public :: reten_name(CHARACTERNUMBER)
    real(DP), allocatable, public :: reten_a(:), reten_n(:), resi_theta(:)
  !end type retentype
  !type :: paratype
    integer(I4), public :: totaln_para = 0
    character(len=VARIABLELENGTH), public :: para_name(CHARACTERNUMBER)
    real(DP), allocatable, public :: sat_hyd_conx(:), sat_hyd_cony(:)
    real(DP), allocatable, public :: sat_hyd_conz(:), spec_stor(:), sat_theta(:)
    character(len=FILELENGTH), public :: para_array_file(PARANUMBER)
    integer(I4), allocatable, public :: parafnum(:)
  !end type paratype
  !type :: inittype
    integer(I4), public :: totaln_init = 0
    character(len=FILELENGTH), public :: in_initial
    real(DP), public :: init_depth = DZERO, rest_time = DZERO
    character(len=VARIABLELENGTH), public :: init_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: init_i(:), init_j(:), init_k(:)
    real(DP), allocatable, public :: init_value(:)
  !end type inittype
  !type :: rechtype
    character(len=TIMELENGTH), public :: rech_unit
    integer(I4), public :: totaln_rech = 0, rech_stepflag = 0
    real(DP), public :: rech_endtime = DINFI, runi_conv = DZERO, rech_multi = DZERO
    character(len=VARIABLELENGTH), public :: rech_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: rech_i(:), rech_j(:)
    real(DP), allocatable, public :: rech_value(:)
  !end type rechtype
  !type :: welltype
    character(len=TIMELENGTH), public :: well_unit
    integer(I4), public :: totaln_well = 0, well_stepflag = 0
    real(DP), public :: well_endtime = DINFI, wuni_conv = DZERO, well_multi = DZERO
    character(len=VARIABLELENGTH), public :: well_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: well_i(:), well_j(:), well_k(:)
    real(DP), allocatable, public :: well_value(:)
  !end type welltype
  !type :: boundtype
    character(len=TIMELENGTH), public :: bound_unit
    integer(I4), public :: totaln_bound = 0, bound_stepflag = 0
    real(DP), public :: bound_endtime = DINFI, bound_multi = DZERO
    character(len=VARIABLELENGTH), public :: bound_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: bound_i(:), bound_j(:), bound_k(:)
    real(DP), allocatable, public :: bound_value(:)
  !end type boundtype
  !type :: surftype
    character(len=TIMELENGTH), public :: surf_unit
    integer(I4), public :: totaln_surf = 0, surf_stepflag = 0
    real(DP), public :: surf_endtime = DINFI, surf_multi = DZERO
    character(len=VARIABLELENGTH), public :: surf_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: surf_i(:), surf_j(:)
    real(DP), allocatable, public :: surf_value(:)
  !end type surftype
  !type :: seeptype
    character(len=TIMELENGTH), public :: seep_unit
    integer(I4), public :: totaln_seep = 0, seep_stepflag = 0
    real(DP), public :: seep_endtime = DINFI, seep_multi = DZERO
    character(len=VARIABLELENGTH), public :: seep_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: seep_i(:), seep_j(:)
  !end type seeptype
  !type :: wlouttype
    integer(I4), public :: totaln_wlout = 0
    character(len=VARIABLELENGTH), public :: wlout_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: wlout_i(:), wlout_j(:), wlout_k(:)
  !end type wlouttype
  !type :: massouttype
    integer(I4), public :: totaln_msout = 0
    character(len=VARIABLELENGTH), public :: msout_class_name(CHARACTERNUMBER)
    integer(I4), allocatable, public :: msout_i(:), msout_j(:), msout_k(:)
  !end type massouttype

  !type :: outflagtype
    integer(I4), public :: outhead_flag = -1, outsrati_flag = -1, outinit_flag = -1
    integer(I4), public :: outwlevel_flag = -1, outmass_flag = -1
    integer(I4), public :: outvelc_flag = -1, outwell_flag = -1
    integer(I4), public :: outrech_flag = -1, outcalgrd_flag = -1
  !end type outflagtype

  ! -- local
  character(len=FILELENGTH) :: inputlist_file, outputlist_file
  character(len=FILELENGTH) :: in_solution, in_grid, in_classification
  character(len=FILELENGTH) :: in_retention, in_parameter
  character(len=FILELENGTH) :: in_recharge, in_well, in_boundary, in_surfwater
  character(len=FILELENGTH) :: in_seepset, in_waterlevelout, in_massbalanceout
  character(len=FILELENGTH) :: out_head, out_srati, out_convergence, out_waterlevel
  character(len=FILELENGTH) :: out_massbalance, out_velocity, out_initial
  character(len=FILELENGTH) :: out_well, out_rech, out_calgrd

  integer(I4), save :: grid_flag
  integer(I4) :: i, j, k, ierr, time_flag
  integer(I4) :: grid_check = 0, grid_xnum = 0, grid_ynum = 0, grid_znum = 0
  real(DP) :: xcorner, ycorner, zcorner
  real(DP) :: xwidth, ywidth, zwidth

  contains

  subroutine read_mainf()
  !*****************************************************************************
  ! read_mainf -- read main file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    character(len=FILELENGTH) :: main_file
    !---------------------------------------------------------------------------
    ! set main file
    main_file = 'msg.main'

    ! open main file
    open(unit=1, file=trim(adjustl(main_file)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error read main file!!"
      stop
    end if

    ! read main file
    read(1, *)
    read(1, *) inputlist_file
    read(1, *)
    read(1, *) outputlist_file
    read(1, *)
    read(1, *) sim_flag
    if (sim_flag == 0 .or. sim_flag == -1) then
      write(*,'(a)') "Simulation type is steady state"
    else if (sim_flag == 1) then
      write(*,'(a)') "Simulation type is unsteady state"
    else
      write(*,'(a)') "Error wrong simulation type"
      stop
    end if

    read(1, *)
    read(1, *) nx, ny, nz
    if (nx == 0 .or. ny == 0 .or. nz == 0) then
      write(*,'(a)') "Error wrong number of cell"
      stop
    end if

    nxyz = nx*ny*nz

    read(1, *)
    read(1, *) start_flag, end_time, calc_unit
    ! -- Convert unit (unit)
      call conv_unit(calc_unit, calc_multi)
    end_time = end_time*calc_multi

    read(1, *)
    read(1, *) init_step, incr_multi, max_step
    init_step = init_step*calc_multi
    max_step = max_step*calc_multi

    read(1, *)
    read(1, *) calcout_flag
!    read(1, *) calcout_flag, calcout_name

    if (calcout_flag == 1) then
      read(1, *) calcout_name
    end if

    ! -- Read inputlist_file (inlistf)
      call read_inlistf()

    ! -- Read outputlist_file (outlistf)
      call read_outlistf()

    return

  end subroutine read_mainf


  subroutine read_inlistf()
  !*****************************************************************************
  ! read_inlistf -- Read input list file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input list file
    open(unit=10, file=trim(adjustl(inputlist_file)), form='formatted',&
        access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error read input list file!!"
      stop
    end if

    ! read input list file
    read(10, *)
    read(10, *)
    read(10, *) in_solution
    ! -- Read input solution file (insolf)
      call read_insolf()

    read(10, *)
    read(10, *) grid_flag, in_grid
    ! -- Read input grid file (ingridf)
      call read_ingridf()

    read(10, *)
    read(10, *) in_classification
    ! -- Read input classification file (inclassf)
      call read_inclassf()

    read(10, *)
    read(10, *) in_retention
    ! -- Read input retention file (inretenf)
      call read_inretenf()

    read(10, *)
    read(10, *) para_flag, in_parameter
    if (para_flag < 0) then
      write(*,'(a)') "Have to set parameter!!"
      stop
    end if
    ! -- Read input parameter file (inparamf)
      call read_inparamf()

    read(10, *)
    read(10, *) init_flag, in_initial
    if (init_flag == 0 .or. init_flag == 1 .or. init_flag == 2) then
      ! -- Read input initial file (inintf)
        call read_inintf()
    else if (init_flag == 3) then
      ! -- Read input initial file (inintf)
        call read_rest()
        rest_time = rest_time*calc_multi
    else
      write(*,'(a)') "Have to set initial conditions!!"
      stop
    end if

    read(10, *)
    read(10, *) rech_flag, in_recharge
    if (rech_flag >= 0) then
      ! -- Read input recharge file (inrechf)
        call read_inrechf()
    else
      write(*,'(a)') "Don't use recharge function."
    end if

    read(10, *)
    read(10, *) well_flag, in_well
    if (well_flag >= 0) then
      ! -- Read input well file (inwellf)
        call read_inwellf()
    else
      write(*,'(a)') "Don't use well function."
    end if

    read(10, *)
    read(10, *) bound_flag, in_boundary
    if (bound_flag >= 0) then
      ! -- Read input boundary file (inboundf)
        call read_inboundf()
    else
      write (*,'(a)') "All boundary conditions are setted close condition."
    end if

    read(10, *)
    read(10, *) surf_flag, in_surfwater
    if (surf_flag >= 0) then
      ! -- Read input surface water file (insurfwf)
        call read_insurfwf()
    else
      write (*,'(a)') "Don't use surface water boundary function."
    end if

    read(10, *)
    read(10, *) seep_flag, in_seepset
    if (seep_flag >= 0) then
      ! -- Read input seepage setting file (inseepset)
        call read_inseepset()
    else
      write (*,'(a)') "Don't use seppage function."
    end if

    read(10, *)
    read(10, *) wlout_flag, in_waterlevelout
    if (wlout_flag >= 1) then
      if (wlout_flag == 1 .or. wlout_flag == 2) then
        ! -- Read input waterlevelout file (inwloutf)
          call read_inwloutf()
      end if
    else
      write(*,'(a)') "Don't use water level output function."
    end if

    read(10, *)
    read(10, *) msout_flag, in_massbalanceout
    if (msout_flag >= 0) then
      if (msout_flag /= 2) then
        ! -- Read input massbalanceout file (inmassoutf)
          call read_inmassoutf()
      end if
    else
      write(*,'(a)') "Don't use massbalance output function."
    end if

    write(*,'(a)') "Read input list file successful!!"

    return

  end subroutine read_inlistf

  subroutine read_outlistf()
  !*****************************************************************************
  ! read_outlistf -- Read output list file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output list file
    open(unit=50, file=trim(adjustl(outputlist_file)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output list file!!"
      stop
    end if

    ! read output list file
    read(50, *)
    read(50, *)
    read(50, *) out_convergence

    ! -- Open output convergence file (outconvf)
      call open_outconvf()

    read(50, *)
    read(50, *) outhead_flag, out_head

    ! -- Open output head file (outheadf)
      call open_outheadf()

    read(50, *)
    read(50, *) outsrati_flag, out_srati

    ! -- Open output saturation file (outsratif)
      call open_outsratif()

    read(50, *)
    read(50, *) outinit_flag, out_initial

    if (outinit_flag > 0) then
      ! -- Open output initial file (outinitf)
        call open_outinitf()
    end if

    read(50, *)
    read(50, *) outwlevel_flag, out_waterlevel

    if (outwlevel_flag > 0) then
      ! -- Open output waterlevel file (outwlevelf)
        call open_outwlevef()
    end if

    read(50, *)
    read(50, *) outmass_flag, out_massbalance

    if (outmass_flag > 0) then
      ! -- Open output massbalance file (outmassf)
        call open_outmassf()
    end if

    read(50, *)
    read(50, *) outvelc_flag, out_velocity

    if (outvelc_flag > 0) then
      ! -- Open output velocity file (outvelcf)
        call open_outvelcf()
    end if

    read(50, *)
    read(50, *) outwell_flag, out_well

    if (outwell_flag > 0) then
      ! -- Open output well pumping results file (outwellf)
        call open_outwellf()
    end if

    read(50, *)
    read(50, *) outrech_flag, out_rech

    if (outrech_flag > 0) then
      ! -- Open output recharge results file (outrechf)
        call open_outrechf()
    end if

    read(50, *)
    read(50, *) outcalgrd_flag, out_calgrd

    if (outcalgrd_flag > 0) then
      ! -- Open output calculation & grid number file (outcalgrdf)
        call open_outcalgrdf()
    end if

    write(*,'(a)') "Read output list file successful!!"

    return

  end subroutine read_outlistf

  subroutine read_insolf()
  !*****************************************************************************
  ! read_insolf -- Read input solution file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input solution file
    open(unit=11, file=trim(adjustl(in_solution)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input solution file!!"
      stop
    end if

    ! read input solution file
    read(11, *)
    ! nonlinear (outer) setting
    read(11, *) maxout_iter
    if (maxout_iter < 0) then
      write(*,'(a)') "Input positive value for outer_maximum"
      stop
    end if
!    read(11, *) criteria
    read(11, *) outche_flag, criteria
    read(11, *) switch_criteria
    read(11, *) newper
    if (switch_criteria >= DONE) then
      write(*,'(a)') "Input a value less than 1.0 for switch criteria"
      stop
    end if
!    outche_flag = 1
    newper_inv = DONE/newper
    ! linear solver (inner) setting
    read(11, *)
    read(11, *) maxinn_iter
    if (maxout_iter < 0) then
      write(*,'(a)') "Input positive value for inner_maximum"
      stop
    end if
    temp_maxinn_iter = maxinn_iter
    read(11, *) errtol
    if (errtol >= DONE) then
      write(*,'(a)') "Input a value less than 1.0 for inner_criteria"
      stop
    end if

    read(11, *) solver_flag
    if (solver_flag > 0 .and. solver_flag < 5) then
      nlevel = 1
      if (solver_flag == 2) then
        read(11, *) sor_omega
        if (sor_omega <= DZERO) then
          write(*,'(a)') "Input positive value for SOR parameter"
          stop
        end if
      else if (solver_flag == 3) then
        read(11, *) sip_alpha
        if (sip_alpha <= DZERO) then
          write(*,'(a)') "Input positive value for SIP parameter"
          stop
        end if
      end if
    else if (solver_flag <= 9) then
      if (solver_flag == 6) then
        read(11, *) sor_omega
        if (sor_omega <= DZERO) then
          write(*,'(a)') "Input positive value for SOR parameter"
          stop
        end if
      else if (solver_flag == 7) then
        read(11, *) sip_alpha
        if (sip_alpha <= DZERO) then
          write(*,'(a)') "Input positive value for SIP parameter"
          stop
        end if
      else
        read(11, *)
      end if
      read(11, *) nlevel, maxvcy_iter, max_swp
    else
      write(*,'(a)') "Input a valid value for solver_flag"
      stop
    end if

    write(*,'(a)') "Read input solution file successful!!"

    return

  end subroutine read_insolf

  subroutine read_ingridf()
  !*****************************************************************************
  ! read_ingridf -- Read input grid file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input grid file
    open(unit=12, file=trim(adjustl(in_grid)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input grid file!!"
      stop
    end if

    allocate(xpoint(nx+1,ny+1), ypoint(nx+1,ny+1))
    xpoint(:,:) = DZERO ; ypoint(:,:) = DZERO
    allocate(zpoint(nx+1,ny+1,nz+1))
    zpoint(:,:,:) = DZERO

    ! read input grid file
    ! read x,y coordinate as 2d array
    read(12, *)
    if (grid_flag == 0) then
      read(12, *) xcorner, xwidth
      do i = 1, nx+1
        xpoint(i,1:ny+1) = xcorner + xwidth*(i-1)
      end do
    else if (grid_flag == 1) then
      do j = ny+1, 1, -1
        read(12, *) (xpoint(i,j), i = 1, nx+1)
      end do
    end if

    read(12, *)
    if (grid_flag == 0) then
      read(12, *) ycorner, ywidth
      do j = 1, ny+1
        ypoint(1:nx+1,j) = ycorner + ywidth*(j-1)
      end do
    else if (grid_flag == 1) then
      do j = ny+1, 1, -1
        read(12, *) (ypoint(i,j), i = 1, nx+1)
      end do
    end if

    ! read z coordinate as 3d array
    read(12, *)
    if (grid_flag == 0) then
      read(12, *) zcorner, zwidth
      do k = 1, nz+1
        zpoint(1:nx+1,1:ny+1,k) = zcorner + zwidth*(k-1)
      end do
    else if (grid_flag == 1) then
      do k = 1, nz+1
        do j = ny+1, 1, -1
          read(12, *) (zpoint(i,j,k), i = 1, nx+1)
        end do
      end do
    end if

    ! -- Check grid location (grdloc)
      call check_grdloc()
    if (grid_check == 0) then
      write(*,'(a)') "Read input grid file successful!!"
    else
      if (grid_check == 1) then
        write(*,'(a,i0,a)') "X direction is reversed at ", grid_xnum, " points"
      end if
      if (grid_check == 2) then
        write(*,'(a,i0,a)') "Y direction is reversed at ", grid_ynum, " points"
      end if
      if (grid_check == 3) then
        write(*,'(a,i0,a)') "Z direction is reversed at ", grid_znum, " points"
      end if
      stop
    end if

    return

  end subroutine read_ingridf

  subroutine read_inclassf()
  !*****************************************************************************
  ! read_inclassf -- Read input classification file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: max_class
    !---------------------------------------------------------------------------
    ! open input classification file
    open(unit=13, file=trim(adjustl(in_classification)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input classification file!!"
      stop
    end if

    ! read input classification file
    read(13, *) totaln_class
    if (totaln_class < 0) then
      write(*,'(a)') "Class number is less than 0. Please check the number"
      stop
    end if

    allocate(class_num(totaln_class))
    class_num(:) = 0
    max_class = 0

    do i = 1, totaln_class
      read(13, *) class_name(i), class_num(i)
      do j = 1, class_num(i)
        read(13, *, iostat=ierr)
        if (ierr /= 0) then
          write(*,'(a)') "Error read input classification file in last line!!"
          stop
        end if
      end do
      if (class_num(i) > max_class) then
        max_class = class_num(i)
      end if
    end do


    allocate(class_i(totaln_class,max_class), class_j(totaln_class,max_class),&
             class_k(totaln_class,max_class))
    class_i(:,:) = 0 ; class_j(:,:) = 0 ; class_k(:,:) = 0

    rewind(13)

    read(13, *) totaln_class
    do i = 1, totaln_class
      read(13, *) class_name(i), class_num(i)
      do j = 1, class_num(i)
        read(13, *) class_i(i,j), class_j(i,j), class_k(i,j)
      end do
    end do

    write(*,'(a)') "Read input classification file successful!!"

    return

  end subroutine read_inclassf

  subroutine read_inretenf()
  !*****************************************************************************
  ! read_inretenf -- Read input retention file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input retention file
    open(unit=14, file=trim(adjustl(in_retention)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input retention file!!"
      stop
    end if

    ! read input retention file
    read(14, *) totaln_reten
    if (totaln_reten < 0) then
      write(*,'(a)') "Retention number is less than 0. Please check the number"
      stop
    end if

    read(14, *)

    allocate(reten_a(totaln_reten), reten_n(totaln_reten),  resi_theta(totaln_reten))
    reten_a(:) = DZERO ; reten_n(:) = DZERO ; resi_theta(:) = DZERO

    do i = 1, totaln_reten
      read(14, *, iostat=ierr) reten_name(i), reten_a(i), reten_n(i), resi_theta(i)
      if (ierr /= 0) then
        write(*,'(a)') "Error read input retention file in last line!!"
        stop
      end if
    end do

    write(*,'(a)') "Read input retention file successful!!"

    return

  end subroutine read_inretenf

  subroutine read_inparamf()
  !*****************************************************************************
  ! read_inparamf -- Read input parameter file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input parameter file
    open(unit=15, file=trim(adjustl(in_parameter)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input parameter file!!"
      stop
    end if


    if (para_flag == 0) then
      ! read input parameter file
      read(15, *) totaln_para
      if (totaln_para < 0) then
        write(*,'(a)') "Parameter number is less than 0. Please check the number"
        stop
      end if

      read(15, *)

      allocate(sat_hyd_conx(totaln_para), sat_hyd_cony(totaln_para))
      allocate(sat_hyd_conz(totaln_para), spec_stor(totaln_para), sat_theta(totaln_para))
      sat_hyd_conx(:) = DZERO ; sat_hyd_cony(:) = DZERO
      sat_hyd_conz(:) = DZERO ; spec_stor(:) = DZERO ; sat_theta(:) = DZERO

      do i = 1, totaln_para
        read(15, *, iostat=ierr) para_name(i), sat_hyd_conx(i), sat_hyd_cony(i),&
                    sat_hyd_conz(i), spec_stor(i), sat_theta(i)
        if (ierr /= 0) then
          write(*,'(a)') "Error read input parameter file in last line!!"
          stop
        end if
      end do

    else if (para_flag == 1) then
      allocate(parafnum(PARANUMBER))
      parafnum(:) = 0
      do i = 1, PARANUMBER
        ! read input parameter array file
        read(15, *)
        read(15, *) para_array_file(i)
        parafnum(i) = 30 + i
        open(unit=parafnum(i), file=trim(adjustl(para_array_file(i))), form='formatted',&
             access='sequential', status='old', action='read', iostat=ierr)
        if (ierr /= 0) then
          write(*,'(a)') "Error open input parameter array file!!"
          stop
        end if
      end do
    end if

    write(*,'(a)') "Read input parameter file successful!!"

    return

  end subroutine read_inparamf

  subroutine read_inintf()
  !*****************************************************************************
  ! read_inintf -- Read input initial file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input initial file
    open(unit=16, file=trim(adjustl(in_initial)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input initial file!!"
      stop
    end if

    if (init_flag == 0) then  !assign classification
      ! read input initial file
      read(16, *) totaln_init
      if (totaln_init < 0) then
        write(*,'(a)') "Initial number is less than 0. Please check the number"
        stop
      end if

      allocate(init_value(totaln_init))
      init_value(:) = DZERO
      do i = 1, totaln_init
        read(16, *, iostat=ierr) init_class_name(i), init_value(i)
        if (ierr /= 0) then
          write(*,'(a)') "Error read input Initial file in last line!!"
          stop
        end if
      end do
    else if (init_flag == 2) then
      read(16, *) init_depth
    else if (init_flag /= 1) then
      write(*,'(a)') "Wrong number of input initial file!!"
      stop
    end if

    write(*,'(a)') "Read input initial file successful!!"

    return

  end subroutine read_inintf

  subroutine read_rest()
  !*****************************************************************************
  ! read_rest -- Read value from restart file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open restart file
!    open(unit=16, file=trim(adjustl(in_initial)), form='formatted',&
!         access='sequential', status='old', action='read', iostat=ierr)
    open(unit=16, file=trim(adjustl(in_initial)), form='unformatted',&
         access='stream', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input restart file!!"
      stop
    end if

    read(16) rest_time

    return

  end subroutine read_rest

  subroutine read_inrechf()
  !*****************************************************************************
  ! read_inrechf -- Read input recharge file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input recharge file
    open(unit=17, file=trim(adjustl(in_recharge)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input recharge file!!"
      stop
    end if

    ! read input recharge file
    read(17, *) rech_unit
    runi_conv = 1.0E-3_DP/HOURSEC
    ! -- Convert unit (unit)
      call conv_unit(rech_unit, rech_multi)

    if (start_flag == 1) then
      time_flag = 0
      do while (time_flag == 0)
        if (rech_flag <=1) then
          read(17, *) totaln_rech, rech_endtime
        else if (rech_flag == 2) then
          read(17, *) rech_endtime
        end if
        rech_endtime = rech_endtime*rech_multi
        if (rech_endtime <= rest_time) then
          if (rech_flag <=1) then
            call skip_time(17, totaln_rech)
          else if (rech_flag == 2) then
            call skip_2darray(17)
          end if
        else
          time_flag = 1
        end if
      end do
    else
      if (rech_flag <=1) then
        read(17, *) totaln_rech, rech_endtime
      else if (rech_flag == 2) then
        read(17, *) rech_endtime
      end if
      rech_endtime = rech_endtime*rech_multi
    end if

    if (rech_endtime > end_time) then
      rech_endtime = end_time
    end if

    allocate(rech_value(totaln_rech))
    rech_value(:) = DZERO

    if (rech_flag == 0) then  !assign classification
      do i = 1, totaln_rech
        read(17, *) rech_class_name(i), rech_value(i)
        rech_value(i) = rech_value(i)*runi_conv
      end do
    else if (rech_flag == 1) then  !assign point
      allocate(rech_i(totaln_rech), rech_j(totaln_rech))
      rech_i(:) = 0 ; rech_j(:) = 0
      do i = 1, totaln_rech
        read(17, *) rech_i(i), rech_j(i), rech_value(i)
        rech_value(i) = rech_value(i)*runi_conv
      end do
    else if (rech_flag /= 2) then
      write(*,'(a)') "Wrong number of input recharge file!!"
      stop
    end if

    rech_stepflag = 0

    write(*,'(a)') "Read input recharge file successful!!"

    return

  end subroutine read_inrechf

  subroutine read_inwellf()
  !*****************************************************************************
  ! read_inwellf -- Read input well file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input well file
    open(unit=18, file=trim(adjustl(in_well)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input well file!!"
      stop
    end if

    ! read input well file
    read(18, *) well_unit
    wuni_conv = DONE/DAYSEC
    ! -- Convert unit (unit)
      call conv_unit(well_unit, well_multi)

    if (start_flag == 1) then
      time_flag = 0
      do while (time_flag == 0)
        read(18, *) totaln_well, well_endtime
        well_endtime = well_endtime*well_multi
        if (well_endtime <= rest_time) then
            call skip_time(18, totaln_well)
        else
          time_flag = 1
        end if
      end do
    else
      read(18, *) totaln_well, well_endtime
      well_endtime = well_endtime*well_multi
    end if

    if (well_endtime > end_time) then
      well_endtime = end_time
    end if


    allocate(well_value(totaln_well))
    well_value(:) = DZERO

    if (well_flag == 0) then  !assign classification
      do i = 1, totaln_well
        read(18, *) well_class_name(i), well_value(i)
        well_value(i) = well_value(i)*wuni_conv
      end do
    else if (well_flag == 1) then  !assign point
      allocate(well_i(totaln_well), well_j(totaln_well), well_k(totaln_well))
      well_i(:) = 0 ; well_j(:) = 0 ; well_k(:) = 0
      do i = 1, totaln_well
        read(18, *) well_i(i), well_j(i), well_k(i), well_value(i)
        well_value(i) = well_value(i)*wuni_conv
      end do
    else
      write(*,'(a)') "Wrong number of input well file!!"
      stop
    end if

    well_stepflag = 0

    write(*,'(a)') "Read input well file successful!!"

    return

  end subroutine read_inwellf

  subroutine read_inboundf()
  !*****************************************************************************
  ! read_inboundf -- Read input boundary file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input boundary file
    open(unit=19, file=trim(adjustl(in_boundary)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input boundary file!!"
      stop
    end if

    ! read input boundary file
    read(19, *) bound_unit
    ! -- Convert unit (unit)
      call conv_unit(bound_unit, bound_multi)

    if (start_flag == 1) then
      time_flag = 0
      do while (time_flag == 0)
        read(19, *) totaln_bound, bound_endtime
        bound_endtime = bound_endtime*bound_multi
        if (bound_endtime <= rest_time) then
            call skip_time(19, totaln_bound)
        else
          time_flag = 1
        end if
      end do
    else
      read(19, *) totaln_bound, bound_endtime
      bound_endtime = bound_endtime*bound_multi
    end if

    if (bound_endtime > end_time) then
      bound_endtime = end_time
    end if


    allocate(bound_value(totaln_bound))
    bound_value(:) = DZERO

    if (bound_flag == 0) then  !assign classification
      do i = 1, totaln_bound
        read(19, *) bound_class_name(i), bound_value(i)
      end do
    else if (bound_flag == 1) then  !assign  grid
      allocate(bound_i(totaln_bound), bound_j(totaln_bound), bound_k(totaln_bound))
      bound_i(:) = 0 ; bound_j(:) = 0 ; bound_k(:) = 0
      do i = 1, totaln_bound
        read(19, *) bound_i(i), bound_j(i), bound_k(i), bound_value(i)
      end do
    else
      write(*,'(a)') "Wrong number of input boundary file!!"
      stop
    end if

    bound_stepflag = 0

    write(*,'(a)') "Read input boundary file successful!!"

    return

  end subroutine read_inboundf

  subroutine read_insurfwf()
  !*****************************************************************************
  ! read_insurfwf -- Read input surface water file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input surface water file
    open(unit=20, file=trim(adjustl(in_surfwater)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input surface water boundary file!!"
      stop
    end if

    ! read input surface water file
    read(20, *) surf_unit
    ! -- Convert unit (unit)
      call conv_unit(surf_unit, surf_multi)

    if (start_flag == 1) then
      time_flag = 0
      do while (time_flag == 0)
        read(20, *) totaln_surf, surf_endtime
        surf_endtime = surf_endtime*surf_multi
        if (surf_endtime <= rest_time) then
            call skip_time(20, totaln_surf)
        else
          time_flag = 1
        end if
      end do
    else
      read(20, *) totaln_surf, surf_endtime
      surf_endtime = surf_endtime*surf_multi
    end if

    if (surf_endtime > end_time) then
      surf_endtime = end_time
    end if


    allocate(surf_value(totaln_surf))
    surf_value(:) = DZERO

    if (surf_flag == 0) then  !assign classification
      do i = 1, totaln_surf
        read(20, *) surf_class_name(i), surf_value(i)
      end do
    else if (surf_flag == 1) then  !assign  grid
      allocate(surf_i(totaln_surf), surf_j(totaln_surf))
      surf_i(:) = 0 ; surf_j(:) = 0
      do i = 1, totaln_surf
        read(20, *) surf_i(i), surf_j(i), surf_value(i)
      end do
    else
      write(*,'(a)') "Wrong number of input surface water boundary file!!"
      stop
    end if

    surf_stepflag = 0

    write(*,'(a)') "Read input surface water file successful!!"

    return

  end subroutine read_insurfwf

  subroutine read_inseepset()
  !*****************************************************************************
  ! read_inseepset -- Read input seepage setting file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input seepage setting file
    open(unit=21, file=trim(adjustl(in_seepset)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input seepage setting file!!"
      stop
    end if

    ! read input seepage setting file
    read(21, *) seep_unit
    ! -- Convert unit (unit)
      call conv_unit(seep_unit, seep_multi)

    if (start_flag == 1) then
      time_flag = 0
      do while (time_flag == 0)
        read(21, *) totaln_seep, seep_endtime
        seep_endtime = seep_endtime*seep_multi
        if (seep_endtime <= rest_time) then
            call skip_time(21, totaln_seep)
        else
          time_flag = 1
        end if
      end do
    else
      read(21, *) totaln_seep, seep_endtime
      seep_endtime = seep_endtime*seep_multi
    end if

    if (seep_endtime > end_time) then
      seep_endtime = end_time
    end if

    if (seep_flag == 0) then  !assign classification
      do i = 1, totaln_seep
        read(21, *) seep_class_name(i)
      end do
    else if (seep_flag == 1) then  !assign  grid
      allocate(seep_i(totaln_seep), seep_j(totaln_seep))
      seep_i(:) = 0 ; seep_j(:) = 0
      do i = 1, totaln_seep
        read(21, *) seep_i(i), seep_j(i)
      end do
    else
      write(*,'(a)') "Wrong number of input seepage setting file!!"
      stop
    end if

    seep_stepflag = 0

    write(*,'(a)') "Read input seepage setting file successful!!"

    return

  end subroutine read_inseepset

  subroutine read_inwloutf()
  !*****************************************************************************
  ! read_inwloutf -- Read input waterlevelout file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input waterlevelout file
    open(unit=51, file=trim(adjustl(in_waterlevelout)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input waterlevelout file!!"
      stop
    end if

    ! read input waterlevelout file
    read(51, *) totaln_wlout
    read(51, *)

!    if (wlout_flag == 0) then  !assign classification
!      do i = 1, totaln_wlout
!        read(51, *) wlout_class_name(i)
!      end do
!    else if (wlout_flag == 1) then  !assign  grid
    if (wlout_flag == 1 .or. wlout_flag == 2) then  !assign  grid
      allocate(wlout_i(totaln_wlout), wlout_j(totaln_wlout), wlout_k(totaln_wlout))
      wlout_i(:) = 0 ; wlout_j(:) = 0 ; wlout_k(:) = 0
      do i = 1, totaln_wlout
        read(51, *) wlout_i(i), wlout_j(i), wlout_k(i)
      end do
    else
      write(*,'(a)') "Wrong number of waterlevelout file!!"
      stop
    end if

    write(*,'(a)') "Read input waterlevelout file successful!!"

    return

  end subroutine read_inwloutf

  subroutine read_inmassoutf()
  !*****************************************************************************
  ! read_inmassoutf -- Read input massbalanceout file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open input massbalanceout file
    open(unit=52, file=trim(adjustl(in_massbalanceout)), form='formatted',&
         access='sequential', status='old', action='read', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open input massbalanceout file!!"
      stop
    end if

    ! read input massbalanceout file
    read(52, *) totaln_msout
    read(52, *)

    if (msout_flag == 0) then  !assign classification
      do i = 1, totaln_msout
        read(52, *) msout_class_name(i)
      end do
    else if (msout_flag == 1) then  !assign  grid
      allocate(msout_i(totaln_msout), msout_j(totaln_msout), msout_k(totaln_msout))
      msout_i(:) = 0 ; msout_j(:) = 0 ; msout_k(:) = 0
      do i = 1, totaln_msout
        read(52, *) msout_i(i), msout_j(i), msout_k(i)
      end do
    else
      write(*,'(a)') "Wrong number of massbalanceout file!!"
      stop
    end if

    write(*,'(a)') "Read input massbalanceout file successful!!"

    return

  end subroutine read_inmassoutf

  subroutine open_outconvf()
  !*****************************************************************************
  ! open_outconvf -- Open output convergence file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output convergence file
    open(unit=101, file=trim(adjustl(out_convergence)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output convergence file!!"
      stop
    end if

    write(*,'(a)') "Open output convergence file successful!!"

    return

  end subroutine open_outconvf

  subroutine open_outheadf()
  !*****************************************************************************
  ! open_outheadf -- Open output head file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output head file
    open(unit=102, file=trim(adjustl(out_head)), form='unformatted',&
         access='stream', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output head file!!"
      stop
    end if

    write(*,'(a)') "Open output head file successful!!"

    return

  end subroutine open_outheadf

  subroutine open_outsratif()
  !*****************************************************************************
  ! open_outsratif -- Open output saturation file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output head file
    open(unit=103, file=trim(adjustl(out_srati)), form='unformatted',&
         access='stream', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output saturation file!!"
      stop
    end if

    write(*,'(a)') "Open output saturation file successful!!"

    return

  end subroutine open_outsratif

  subroutine open_outinitf()
  !*****************************************************************************
  ! open_outinitf -- Open output initial file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output initial file
    open(unit=104, file=trim(adjustl(out_initial)), form='unformatted',&
         access='stream', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output initial file!!"
      stop
    end if

    write(*,'(a)') "Open output initial file successful!!"

    return

  end subroutine open_outinitf

  subroutine open_outwlevef()
  !*****************************************************************************
  ! open_outwlevef -- Open output waterlevel file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output waterlevel file
    open(unit=105, file=trim(adjustl(out_waterlevel)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output waterlevel file!!"
      stop
    end if

    write(*,'(a)') "Open output waterlevel file successful!!"

    return

  end subroutine open_outwlevef

  subroutine open_outmassf()
  !*****************************************************************************
  ! open_outmassf -- Open output massbalance file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output massbalance file
    open(unit=106, file=trim(adjustl(out_massbalance)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output massbalance file!!"
      stop
    end if

    write(*,'(a)') "Open output massbalance file successful!!"

    return

  end subroutine open_outmassf

  subroutine open_outvelcf()
  !*****************************************************************************
  ! open_outvelcf -- Open output velocity file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output velocity file
    open(unit=107, file=trim(adjustl(out_velocity)), form='unformatted',&
         access='stream', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output velocity file!!"
      stop
    end if

    write(*,'(a)') "Open output velocity file successful!!"

    return

  end subroutine open_outvelcf

  subroutine open_outwellf()
  !*****************************************************************************
  ! open_outwellf -- Open output well pumping results file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open well results file
    open(unit=108, file=trim(adjustl(out_well)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output well results file!!"
      stop
    end if

    write(*,'(a)') "Open output well results file successful!!"

    return

  end subroutine open_outwellf

  subroutine open_outrechf()
  !*****************************************************************************
  ! open_outrechf -- Open output recharge results file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open recharge results file
    open(unit=109, file=trim(adjustl(out_rech)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output recharge results file!!"
      stop
    end if

    write(*,'(a)') "Open output recharge results file successful!!"

    return

  end subroutine open_outrechf

  subroutine open_outcalgrdf()
  !*****************************************************************************
  ! open_outcalgrdf -- Open output calculation & grid number file
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! open output calculation & grid number file
    open(unit=200, file=trim(adjustl(out_calgrd)), form='formatted',&
         access='sequential', status='replace', action='write', iostat=ierr)
    if (ierr /= 0) then
      write(*,'(a)') "Error open output calculation & grid number file!!"
      stop
    end if

    write(*,'(a)') "Open output calculation & grid number file successful!!"

    return

  end subroutine open_outcalgrdf

  subroutine check_grdloc()
  !*****************************************************************************
  ! check_grdloc -- Check grid location
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    !---------------------------------------------------------------------------
    do j = 1, ny+1
      do i = 2, nx+1
        if (xpoint(i,j) < xpoint(i-1,j)) then
          grid_check = 1
          grid_xnum = grid_xnum + 1
        end if
      end do
    end do

    do i = 1, nx+1
      do j = 2, ny+1
        if (ypoint(i,j) < ypoint(i,j-1)) then
          grid_check = 2
          grid_ynum = grid_ynum + 1
        end if
      end do
    end do

    do i = 1, nx+1
      do j = 1, ny+1
        do k = 2, nz+1
          if (zpoint(i,j,k) > zpoint(i,j,k-1)) then
            grid_check = 3
            grid_znum = grid_znum + 1
          end if
        end do
      end do
    end do

    return

  end subroutine check_grdloc

  subroutine skip_time(file_num, targetn)
  !*****************************************************************************
  ! skip_time -- Skip time series input file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: file_num, targetn
    ! -- local
    !---------------------------------------------------------------------------
    do i = 1, targetn
      read(file_num, *)
    end do

    return

  end subroutine skip_time

  subroutine skip_2darray(file_num)
  !*****************************************************************************
  ! skip_2darray -- Read skip time series input file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: file_num
    ! -- local
    real(DP) :: dummy
    !---------------------------------------------------------------------------
    do j = 1, ny
      read(file_num, *) (dummy, i = 1, nx)
    end do

    return

  end subroutine skip_2darray

  subroutine conv_unit(time_char, time_conv)
  !*****************************************************************************
  ! conv_unit -- Convert unit
  !*****************************************************************************
    ! -- modules

    ! -- inout
    character(len=*), intent(in) :: time_char
    real(DP), intent(out) :: time_conv
    ! -- local
    !---------------------------------------------------------------------------
    if (trim(adjustl(time_char)) == "SEC") then
      time_conv = DONE
    else if (trim(adjustl(time_char)) == "MIN") then
      time_conv = MINUTESEC
    else if (trim(adjustl(time_char)) == "HOUR") then
      time_conv = HOURSEC
    else if (trim(adjustl(time_char)) == "DAY") then
      time_conv = DAYSEC
    else if (trim(adjustl(time_char)) == "YEAR") then
      time_conv = YEARSEC
    end if

    return

  end subroutine conv_unit

end module read_inputfile
