module time_module
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE, DTWO, DHALF, DSMAL, DNOVAL
  use read_inputfile, only: sim_flag, solver_flag, init_step, max_step,&
                            incr_multi, end_time, totaln_rech, rech_flag,&
                            rech_stepflag, rech_multi, rech_endtime, runi_conv,&
                            rech_class_name, rech_i, rech_j, rech_value,&
                            totaln_well, well_flag, well_stepflag, well_multi,&
                            well_endtime, wuni_conv, well_class_name, well_i,&
                            well_j, well_k, well_value, totaln_bound,&
                            bound_flag, bound_stepflag, bound_multi,&
                            bound_endtime, bound_class_name, bound_i, bound_j,&
                            bound_k, bound_value, totaln_surf, surf_flag,&
                            surf_stepflag, surf_multi, surf_endtime,&
                            surf_class_name, surf_i, surf_j, surf_value,&
                            start_flag, rest_time, calc_multi, maxinn_iter,&
                            temp_maxinn_iter, outhead_flag, outsrati_flag
  use set_classification, only: ncalc, ncals, calc2cell, calc2cals, calc_end,&
                                cals2surf
  use make_cell, only: vol_c, area, surfelev, dis, cell_botelev, cell_centelev
  use set_condition, only: set_class2cell2, set_class2surf, set_point2cell,&
                           set_point2surf, read_2darray, g_index
  use set_mg_condition, only: set_mgareave, set_mgsum, set_mgbound, set_mgvolave
  use set_calculation, only: cell_ksx, cell_ksy, cell_ksz, cell_sat, cell_res,&
                             cell_head, cell_rech, cell_well, cell_bound,&
                             ghost_head, cell_surf, surf_head, rech_cflag,&
                             well_cflag, bound_cflag, ghost_bound, surf_cflag,&
                             seep_cflag, current_t
  use calc_parameter, only: calc_srati
  use allocate_module, only: head_old, head_new, srati_old, srati_new,&
                             surf_old, rech_old, well_old, calc_rech, calc_well

  implicit none
  private
  public :: update_tstep
  integer(I4), public :: conv_flag = 0, form_swich = 0, timestep_num = 0
  real(DP), public :: delt = DZERO, now_time = DZERO

  ! -- local
  integer(I4) :: i, c, s, g, n, ierr, totaln_ghost

  contains

  subroutine update_tstep()
  !*****************************************************************************
  ! update_tstep -- Update time step
  !*****************************************************************************
    ! -- module

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calcultae next time step (nextst)
      call calc_nextst()

    if (conv_flag == 1 .and. sim_flag == 1) then
      ! -- Set time series condition at current time (tseries)
        call set_tseries()
    else if (current_t + delt > end_time) then
      delt = end_time - current_t
    end if

    if (delt < DSMAL .and. sim_flag /= -1) then
      write(*,'(a)') "Time Step is too small!!"
      stop
    else if (sim_flag /= -1) then
      current_t = current_t + delt
    else if (current_t == DZERO .and. sim_flag == -1) then
      current_t = current_t + delt
    end if

    now_time = current_t/calc_multi

    if (sim_flag /= -1) then
    ! -- Change the volume rate using time step & cell volume (volrate)
      call change_volrate()
    end if

    return

  end subroutine update_tstep

  subroutine calc_nextst()
  !*****************************************************************************
  ! calc_nextst -- Calcultae next time step
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: cell_srati
    !---------------------------------------------------------------------------
    if (current_t == DZERO) then
      timestep_num = 1
      if (sim_flag == -1) then
        delt = DNOVAL
        maxinn_iter = 100
      else
        delt = init_step
        maxinn_iter = temp_maxinn_iter
      end if
      do i = 1, ncalc
        s = calc2cals(i)
        if (s /= 0) then
          if (surf_cflag(s) == 1) then
            surf_head(s) = cell_surf(s)
          else
            surf_head(s) = cell_head(i)
          end if
          calc_rech(s) = cell_rech(s) ; rech_old(s) = cell_rech(s)
        end if
        ! -- Calculate saturation ratio (srati)
          call calc_srati(i, DZERO, cell_head(i), cell_srati)
        head_old(i) = cell_head(i) ; head_new(i) = cell_head(i)
        srati_old(i) = cell_srati ; srati_new(i) = cell_srati
        calc_well(i) = cell_well(i) ; well_old(i) = cell_well(i)
      end do
      if (start_flag == 1) then
        current_t = rest_time*calc_multi
      end if
      now_time = current_t/calc_multi
      if (outhead_flag > 0) then
        write(102) now_time, (head_new(i), i = 1, calc_end(1))
      end if
      if (outsrati_flag > 0) then
        write(103) now_time, (srati_new(i), i = 1, calc_end(1))
      end if
    else if (sim_flag >= 0) then
      if (conv_flag == 1) then
        timestep_num = timestep_num + 1
        if (rech_stepflag == 1 .or. well_stepflag == 1 .or. bound_stepflag ==1 .or. surf_stepflag ==1) then
          delt = delt*DHALF
        else
          delt = delt*incr_multi
        end if
        head_old(:) = head_new(:) ; srati_old(:) = srati_new(:)
        surf_old(:) = surf_head(:) ; rech_old(:) = cell_rech(:)
        well_old(:) = cell_well(:)
      else
        current_t = current_t - delt
        delt = delt*DHALF
        head_new(:) = head_old(:) ; srati_new(:) = srati_old(:)
        surf_head(:) = surf_old(:) ; cell_rech(:) = rech_old(:)
        cell_well(:) = well_old(:)
        if (sim_flag == 1) then
          if (rech_stepflag == 1) then
            rech_stepflag = 0
          end if
          if (well_stepflag == 1) then
            well_stepflag = 0
          end if
          if (bound_stepflag == 1) then
            bound_stepflag = 0
          end if
          if (surf_stepflag == 1) then
            surf_stepflag = 0
          end if
        end if
      end if
      if (solver_flag >= 5) then
        ! -- Set mg condition volume average (mgvolave)
          call set_mgvolave(head_new)
      end if
    else if (sim_flag == -1) then
      head_old(:) = head_new(:) ; srati_old(:) = srati_new(:)
    end if

    return

  end subroutine calc_nextst

  subroutine set_tseries()
  !*****************************************************************************
  ! set_tseries -- Set time series condition at current time
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: min_step
    !---------------------------------------------------------------------------
    if (rech_stepflag == 1) then
      if (rech_flag <= 1) then
        read(17, *, iostat=ierr) totaln_rech, rech_endtime
      else if (rech_flag == 2) then
        read(17, *, iostat=ierr) rech_endtime
      end if
      rech_endtime = rech_endtime*rech_multi
      if (ierr /= 0) then
        write(*,'(a)') "Read final time step recharge file!!"
        rech_stepflag = 0
        rech_endtime = end_time
        ! -- Reset cell value & flag (valflag)
          call reset_valflag(ncals, cell_rech, rech_cflag)
      end if
    end if

    if (well_stepflag == 1) then
      read(18, *, iostat=ierr) totaln_well, well_endtime
      well_endtime = well_endtime*well_multi
      if (ierr /= 0) then
        write(*,'(a)') "Read final time step well file!!"
        well_stepflag = 0
        well_endtime = end_time
        ! -- Reset cell value & flag (valflag)
          call reset_valflag(ncalc, cell_well, well_cflag)
      end if
    end if

    if (bound_stepflag == 1) then
      read(19, *, iostat=ierr) totaln_bound, bound_endtime
      bound_endtime = bound_endtime*bound_multi
      if (ierr /= 0) then
        write(*,'(a)') "Read final time step boundary file!!"
        bound_stepflag = 0
        bound_endtime = end_time
      end if
    end if

    if (surf_stepflag == 1) then
      read(20, *, iostat=ierr) totaln_surf, surf_endtime
      surf_endtime = surf_endtime*surf_multi
      if (ierr /= 0) then
        write(*,'(a)') "Read final time step surface water boundary file!!"
        surf_stepflag = 0
        surf_endtime = end_time
      end if
    end if

    min_step = min(rech_endtime, well_endtime, bound_endtime, surf_endtime)
    if (min_step == current_t) then
      delt = delt
    else if (min_step < current_t + delt .and. current_t + delt < end_time) then
      delt = min_step - current_t
    else if (current_t + delt > end_time) then
      delt = end_time - current_t
    end if

    if (delt > max_step) then
      delt = max_step
    end if

    if (rech_stepflag == 1) then
      ! deallocate & reallocate
      deallocate(rech_value)
      allocate(rech_value(totaln_rech))
      rech_value(:) = DZERO
      if (rech_flag == 0) then
      ! -- Read time series classification input file (tsclassf)
        call read_tsclassf(17, totaln_rech, runi_conv, rech_class_name, rech_value)
        call set_class2surf(totaln_rech, rech_class_name, rech_value, cell_rech, rech_cflag)
      else if (rech_flag == 1) then
        ! deallocate & reallocate
        deallocate(rech_i, rech_j)
        allocate(rech_i(totaln_rech), rech_j(totaln_rech))
        rech_i(:) = 0 ; rech_j(:) = 0
      ! -- Read time series point input file (tspointf)
        call read_tspointfsurf(17, totaln_rech, runi_conv, rech_i, rech_j, rech_value)
        call set_point2surf(totaln_rech, rech_i, rech_j, rech_value, cell_rech, rech_cflag)
      else if (rech_flag == 2) then
      ! -- Read time series array input file (2darray)
        call read_2darray(17, runi_conv, cell_rech)
      end if
      rech_stepflag = 0
      write(101,'(a,f15.3)') "Changed recharge condition at ", now_time
      if (solver_flag >= 5) then
      ! -- Set mg condition area average (mgareave)
        call set_mgareave(cell_rech, rech_cflag)
      end if
    else if (rech_endtime == current_t + delt) then
      rech_stepflag = 1
    end if

    if (well_stepflag == 1) then
      ! deallocate & reallocate
      deallocate(well_value)
      allocate(well_value(totaln_well))
      well_value(:) = DZERO
      if (well_flag == 0) then
      ! -- Read time series classification input file (tsclassf)
        call read_tsclassf(18, totaln_well, wuni_conv, well_class_name, well_value)
        call set_class2cell2(totaln_well, well_class_name, well_value, cell_well, well_cflag)
      else if (well_flag == 1) then
        ! deallocate & reallocate
        deallocate(well_i, well_j, well_k)
        allocate(well_i(totaln_well), well_j(totaln_well), well_k(totaln_well))
        well_i(:) = 0 ; well_j(:) = 0 ; well_k(:) = 0
      ! -- Read time series point input file (tspointf)
        call read_tspointf(18, totaln_well, wuni_conv, well_i, well_j, well_k, well_value)
        call set_point2cell(totaln_well, well_i, well_j, well_k, well_value, cell_well, well_cflag)
      end if
      well_stepflag = 0
      write(101,'(a,f15.3)') "Changed well condition at ", now_time
      if (solver_flag >= 5) then
      ! -- Set mg condition summation (mgsum)
        call set_mgsum(cell_well, well_cflag)
      end if
    else if (well_endtime == current_t + delt) then
      well_stepflag = 1
    end if

    if (bound_stepflag == 1) then
      ! deallocate & reallocate
      deallocate(bound_value)
      allocate(bound_value(totaln_bound))
      bound_value(:) = DZERO
      if (bound_flag == 0) then
      ! -- Read time series classification input file (tsclassf)
        call read_tsclassf(19, totaln_bound, DONE, bound_class_name, bound_value)
        call set_class2cell2(totaln_bound, bound_class_name, bound_value, cell_bound, bound_cflag)
      else if (bound_flag == 1) then
        ! deallocate & reallocate
        deallocate(bound_i, bound_j, bound_k)
        allocate(bound_i(totaln_bound), bound_j(totaln_bound), bound_k(totaln_bound))
        bound_i(:) = 0 ; bound_j(:) = 0 ; bound_k(:) = 0
      ! -- Read time series point input file (tspointf)
        call read_tspointf(19, totaln_bound, DONE, bound_i, bound_j, bound_k, bound_value)
        call set_point2cell(totaln_bound, bound_i, bound_j, bound_k, bound_value, cell_bound, bound_cflag)
      end if
      bound_stepflag = 0
      write(101,'(a,f15.3)') "Changed boundary condition at ", now_time
      if (solver_flag >= 5) then ! if multi grid method is used
      ! -- Set mg boundary condition (mgbound)
        call set_mgbound(head_new, cell_bound, bound_cflag, ghost_bound, totaln_ghost)
      end if
      do i = 1, ncalc
        if (bound_cflag(i) == 1) then
          cell_head(i) = cell_bound(i)
        else if (bound_cflag(i) == 2) then
          do g = g_index(i-1)+1, g_index(i)
            ghost_head(g) = ghost_bound(i)
          end do
        end if
      end do
    else if (bound_endtime == current_t + delt) then
      bound_stepflag = 1
    end if

    if (surf_stepflag == 1) then
      ! deallocate & reallocate
      deallocate(surf_value)
      allocate(surf_value(totaln_surf))
      surf_value(:) = DZERO
      if (surf_flag == 0) then
      ! -- Read time series classification input file (tsclassf)
        call read_tsclassf(20, totaln_surf, DONE, surf_class_name, surf_value)
        call set_class2surf(totaln_surf, surf_class_name, surf_value, cell_surf, surf_cflag)
      else if (surf_flag == 1) then
        ! deallocate & reallocate
        deallocate(surf_i, surf_j)
        allocate(surf_i(totaln_surf), surf_j(totaln_surf))
        surf_i(:) = 0 ; surf_j(:) = 0
      ! -- Read time series point input file (tspointf)
        call read_tspointfsurf(20, totaln_surf, DONE, surf_i, surf_j, surf_value)
        call set_point2surf(totaln_surf, surf_i, surf_j, surf_value, cell_surf, surf_cflag)
      end if
      surf_stepflag = 0
      write(101,'(a,f15.3)') "Changed surface condition at ", now_time
      if (solver_flag >= 5) then ! if multi grid method is used
      ! -- Set mg condition area average (mgareave)
        call set_mgareave(cell_surf, surf_cflag)
      end if
      do i = 1, ncals
        if (surf_cflag(i) == 1) then
          surf_head(i) = cell_surf(i)
        end if
      end do
    else if (surf_endtime == current_t + delt) then
      surf_stepflag = 1
    end if

    return

  end subroutine set_tseries

  subroutine change_volrate()
  !*****************************************************************************
  ! change_volrate -- Change the volume rate using time step & cell volume
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: ss
    real(DP) :: head_rati
    !---------------------------------------------------------------------------
    do i = 1, ncalc
      if (bound_cflag(i) /= 1) then
        c = calc2cell(i)
        s = calc2cals(i)
        if (s /= 0) then
          ss = cals2surf(s)
          calc_rech(s) = cell_rech(s)
        end if
        calc_well(i) = cell_well(i)

        if (s == 0 .and. cell_well(i) < DZERO) then !only well pumping
          if (head_new(i) < cell_botelev(c)) then
            calc_well(i) = DZERO
          else if (head_new(i) < cell_centelev(c)) then
            head_rati = (cell_centelev(c)-head_new(i))/(cell_centelev(c)-cell_botelev(c))
            calc_well(i) = calc_well(i)*(DONE-head_rati)
          end if
        else if (s /= 0) then
          if (surf_head(s) > surfelev(ss) .and. rech_cflag(s) == 1) then !exist surface water
            surf_head(s) = surf_head(s) + cell_rech(s)*delt
            if (surf_head(s) < surfelev(ss)) then !below groundlevel
              calc_rech(s) = (surf_head(s) - surfelev(ss)) / delt
              surf_head(s) = surfelev(ss)
            else
              calc_rech(s) = DZERO
            end if
          end if
        end if
      end if
    end do

    return

  end subroutine change_volrate

  subroutine read_tsclassf(file_num, targetn, target_unit, target_name, target_value)
  !*****************************************************************************
  ! read_tsclassf -- Read time series classification input file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: file_num, targetn
    real(DP), intent(in) :: target_unit
    character(len=*), intent(out) :: target_name(targetn)
    real(DP), intent(out) :: target_value(targetn)
    ! -- local

    !---------------------------------------------------------------------------
    do i = 1, targetn
      read(file_num, *) target_name(i), target_value(i)
      target_value(i) = target_value(i)*target_unit
    end do

    return

  end subroutine read_tsclassf

  subroutine read_tspointf(file_num, targetn, target_unit, p_i, p_j, p_k, target_value)
  !*****************************************************************************
  ! read_tspointf -- Read time series point input file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: file_num, targetn
    real(DP),  intent(in) :: target_unit
    integer(I4), intent(out) :: p_i(targetn), p_j(targetn), p_k(targetn)
    real(DP), intent(out) :: target_value(targetn)
    ! -- local
    integer(I4) :: i
    !---------------------------------------------------------------------------
    do i = 1, targetn
      read(file_num, *) p_i(i), p_j(i), p_k(i), target_value(i)
      target_value(i) = target_value(i)*target_unit
    end do

    return

  end subroutine read_tspointf

  subroutine read_tspointfsurf(file_num, targetn, target_unit, p_i, p_j, target_value)
  !*****************************************************************************
  ! read_tspointfsurf -- Read time series point input file for xy plane data
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: file_num, targetn
    real(DP),  intent(in) :: target_unit
    integer(I4), intent(out) :: p_i(targetn), p_j(targetn)
    real(DP), intent(out) :: target_value(targetn)
    ! -- local
    integer(I4) :: i
    !---------------------------------------------------------------------------
    do i = 1, targetn
      read(file_num, *) p_i(i), p_j(i), target_value(i)
      target_value(i) = target_value(i)*target_unit
    end do

    return

  end subroutine read_tspointfsurf

  subroutine reset_valflag(targetn, target_value, target_cflag)
  !*****************************************************************************
  ! reset_valflag -- Reset cell value & flag
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    real(DP), intent(inout) :: target_value(:)
    integer(I4), intent(inout) :: target_cflag(:)
    ! -- local
    integer(I4) :: i
    !---------------------------------------------------------------------------
    do i = 1, targetn
      target_value(i) = DZERO
      target_cflag(i) = 0
    end do

    return

  end subroutine reset_valflag

end module time_module
