module calc_simulation
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DTWO
  use set_classification, only: calc_end, ncalc
  use set_condition, only: off_index, off_row
  use set_calculation, only: bound_cflag
  use allocate_module, only: dmat, lumat

  implicit none
  private
  public :: calc_l2norm2, calc_resi, calc_resl2norm

  ! -- local
  integer(I4) :: i, j, k

  contains

  subroutine calc_l2norm2(l2level, inbr, l2norm2)
  !*****************************************************************************
  ! calc_l2norm2 -- Calculate l2 norm square
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: l2level
    real(DP), intent(in) :: inbr(:)
    real(DP), intent(out) :: l2norm2
    ! -- local
    integer(I4) :: l2_sta, l2_end, bi
    !---------------------------------------------------------------------------
    l2_sta = calc_end(l2level-1) + 1
    l2_end = calc_end(l2level)

    l2norm2 = DZERO

    do i = l2_sta, l2_end
      if (bound_cflag(i) /= 1) then
        l2norm2 = l2norm2 + inbr(i)*inbr(i)
      end if
    end do

    return

  end subroutine calc_l2norm2

  subroutine calc_resi(rlevel, estx, curb, resb)
  !*****************************************************************************
  ! calc_resi -- Calculate residual
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: rlevel
    real(DP), intent(in) :: estx(:), curb(:)
    real(DP), intent(out) :: resb(ncalc)
    ! -- local
    real(DP) :: axsum
    integer(I4) :: res_sta, res_end, xi
    !---------------------------------------------------------------------------
    res_sta = calc_end(rlevel-1) + 1
    res_end = calc_end(rlevel)

    do i = res_sta, res_end
!      xi = i - res_sta + 1
      axsum = DZERO
!      axsum = dmat(i)*estx(xi)
      axsum = dmat(i)*estx(i)
      do k = off_index(i-1)+1, off_index(i)
!        j = off_row(k) - res_sta + 1
        j = off_row(k)
        axsum = axsum + lumat(k)*estx(j)
      end do
!      resb(i) = curb(xi) - axsum
      resb(i) = curb(i) - axsum
    end do

    return

  end subroutine calc_resi

  subroutine calc_resl2norm(res2_sta, res2_end, res2x, res2b, resl2)
  !*****************************************************************************
  ! calc_resl2norm -- Calculate residual & l2 norm square
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: res2_sta, res2_end
    real(DP), intent(in) :: res2x(:), res2b(:)
    real(DP), intent(out) :: resl2
    ! -- local
    real(DP) :: res2sum
    !---------------------------------------------------------------------------
    resl2 = DZERO

    do i = res2_sta, res2_end
      if (bound_cflag(i) /= 1) then
        res2sum = DZERO
        res2sum = dmat(i)*res2x(i)
        do k = off_index(i-1)+1, off_index(i)
          j = off_row(k)
          res2sum = res2sum + lumat(k)*res2x(j)
        end do
        resl2 = resl2 + (res2b(i) - res2sum)**DTWO
      end if
    end do

    return

  end subroutine calc_resl2norm

end module calc_simulation
