module check_simulation
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO
  use read_inputfile, only: end_time, errtol, sim_flag, rech_stepflag,&
                            well_stepflag, bound_stepflag, surf_stepflag
  use set_classification, only: nunknow
  use set_calculation, only: current_t, bound_cflag
  use time_module, only: conv_flag

  implicit none
  private
  public :: check_insol, check_abserrmax, check_outtiming, check_lastts
  integer(I4), public :: write_flag = 0, lasttime_flag =0

  ! -- local

  contains

  subroutine check_insol(bsum, rsum)
    !***************************************************************************
    ! check_insol -- Check inner solution
    !***************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: bsum, rsum
    ! -- local
    real(DP) :: rerr
!    integer(I4) :: i
    !---------------------------------------------------------------------------
    rerr = sqrt(rsum/bsum)

    if (rerr <= errtol) then
      conv_flag = 1
    else
      conv_flag = 0
    end if

    return

  end subroutine check_insol

  subroutine check_abserrmax(x_new, x_pre, chmax, xmax, nunmax)
    !***************************************************************************
    ! check_abserrmax -- Check absolute error max norm
    !***************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: x_new(:), x_pre(:)
    real(DP), intent(out) :: chmax, xmax
    integer(I4), intent(out) :: nunmax
    ! -- local
    real(DP) :: ach, aval, bch, bval, xdif, xdif_abs, x_abs
    integer(I4) :: i, nun
    !---------------------------------------------------------------------------
    aval = DZERO
    bval = DZERO
    bch = DZERO
    ach = DZERO
    do i = 1, nunknow
      if (x_new(i) /= x_new(i)) then
        bch = DZERO ; bval = x_new(i) ; nun = i
        exit
      end if
      x_abs = abs(x_new(i))
      xdif = x_new(i) - x_pre(i)
      xdif_abs = abs(xdif)
      if (xdif_abs >= ach) then
        bch = xdif
        ach = xdif_abs
        nun = i
      end if
      if (x_abs >= aval) then
        aval = x_abs
        bval = x_new(i)
      end if
    end do

    chmax = bch
    xmax = bval
    nunmax = nun

    return

  end subroutine check_abserrmax

  subroutine check_outtiming()
    !***************************************************************************
    ! check_outtiming -- check output timing
    !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: step_flag
    !---------------------------------------------------------------------------
    step_flag = rech_stepflag + well_stepflag + bound_stepflag + surf_stepflag
    if (step_flag > 0) then
      write_flag = 1
    else if (current_t >= end_time) then
      write_flag = 1
    else if (conv_flag == 1 .and. sim_flag /= 1) then
      write_flag = 1
    else
      write_flag = 0
    end if

    return

  end subroutine check_outtiming

  subroutine check_lastts()
    !***************************************************************************
    ! check_lastts -- Check last time step
    !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (current_t >= end_time) then
      lasttime_flag = 1
    else if (conv_flag == 1 .and. sim_flag == -1) then
      lasttime_flag = 1
    else
      lasttime_flag = 0
    end if

    return

  end subroutine check_lastts

end module check_simulation
