module set_calculation
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DNOVAL, FACE
  use read_inputfile, only: solver_flag, totaln_reten, reten_name, reten_a,&
                            reten_n, resi_theta, para_flag, totaln_para,&
                            para_name, para_array_file, sat_hyd_conx,&
                            sat_hyd_cony, sat_hyd_conz, spec_stor, sat_theta,&
                            init_flag, totaln_init, init_class_name, init_value,&
                            rech_flag, totaln_rech, rech_class_name,&
                            rech_value, rech_i, rech_j, runi_conv,&
                            well_flag, totaln_well, well_class_name,&
                            well_value, well_i, well_j, well_k,&
                            bound_flag, totaln_bound, bound_class_name,&
                            bound_value, bound_i, bound_j, bound_k,&
                            surf_flag, totaln_surf, surf_class_name,&
                            surf_value, surf_i, surf_j, rest_time,&
                            seep_flag, totaln_seep, seep_class_name,&
                            seep_i, seep_j
  use set_classification, only: set_class, ncalc, ncals, calc2cals, calc2cell,&
                                nunknow
  use make_cell, only: make_cellinfo, vol_c, area
  use set_condition, only: set_class2cell1, set_class2cell2, set_class2surf,&
                           set_class2seep, set_point2cell, set_point2surf,&
                           set_point2seep, set_connect, set_glevel,&
                          read_3darray, read_2darray, g_index
  use set_mg_condition, only: set_mgvol, set_mgvolave, set_mgare, set_mgareave, set_mgsum,&
                              set_mgbound, set_mgseep, set_proinfo, set_resinfo

  implicit none
  private
  public :: prepare_calc
  real(DP), public :: current_t
  real(DP), allocatable, public :: cell_reta(:), cell_retn(:), cell_res(:)
  real(DP), allocatable, public :: cell_ksx(:), cell_ksy(:), cell_ksz(:),&
                                   cell_ss(:), cell_sat(:)
  real(DP), allocatable, public :: cell_rech(:), cell_well(:), cell_init(:), cell_bound(:)
  integer(I4), allocatable, public :: rech_cflag(:), well_cflag(:)
  integer(I4), allocatable, public :: bound_cflag(:), surf_cflag(:), seep_cflag(:)
  real(DP), allocatable, public :: cell_head(:), ghost_bound(:), ghost_head(:)
  real(DP), allocatable, public :: cell_surf(:), surf_head(:)

  ! -- local
  integer(I4) :: i, s, g, check_flag, ng, nghost

  contains

  subroutine prepare_calc
  !***********************************************************************
  ! prepare_calc -- Prepare calculation
  !***********************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! set current time
    current_t = DZERO

    ! -- Set classification & calculation cell (class)
      call set_class()

    ! -- Make cell information (cellinfo)
      call make_cellinfo()

    ! allocate
    allocate(cell_reta(ncalc), cell_retn(ncalc), cell_res(ncalc))
    cell_reta(:) = DNOVAL ; cell_retn(:) = DNOVAL ; cell_res(:) = DNOVAL

    ! -- Set retention to cell using cell classification (class2cell)
      call set_class2cell1(totaln_reten, reten_name, reten_a, cell_reta)
      call set_class2cell1(totaln_reten, reten_name, reten_n, cell_retn)
      call set_class2cell1(totaln_reten, reten_name, resi_theta, cell_res)

    ! allocate
    allocate(cell_ksx(ncalc), cell_ksy(ncalc), cell_ksz(ncalc))
    allocate(cell_ss(ncalc), cell_sat(ncalc))
    cell_ksx(:) = DNOVAL ; cell_ksy(:) = DNOVAL ; cell_ksz(:) = DNOVAL
    cell_ss(:) = DNOVAL ; cell_sat(:)= DNOVAL

    if (para_flag == 0) then  !classification
      call set_class2cell1(totaln_para, para_name, sat_hyd_conx, cell_ksx)
      call set_class2cell1(totaln_para, para_name, sat_hyd_cony, cell_ksy)
      call set_class2cell1(totaln_para, para_name, sat_hyd_conz, cell_ksz)
      call set_class2cell1(totaln_para, para_name, spec_stor, cell_ss)
      call set_class2cell1(totaln_para, para_name, sat_theta, cell_sat)
    else if (para_flag == 1) then  !array
      call read_3darray(31, cell_ksx)
      call read_3darray(32, cell_ksy)
      call read_3darray(33, cell_ksz)
      call read_3darray(34, cell_ss)
      call read_3darray(35, cell_sat)
    end if

    ! allocate
    allocate(cell_init(ncalc))
    cell_init(:) = DNOVAL

    ! set initial to cell
    if (init_flag == 0) then  !classification
      call set_class2cell1(totaln_init, init_class_name, init_value, cell_init)
    else if (init_flag == 1) then  !array
      call read_3darray(16, cell_init)
    else if (init_flag == 2) then  !groundlevel
      call set_glevel(cell_init)
    else if (init_flag == 3) then  !restart
      rewind(16)
      read(16) rest_time, (cell_init(i), i = 1, nunknow)
      close(16)
    end if

    ! allocate
    allocate(cell_rech(ncals))
    allocate(rech_cflag(ncals))
    cell_rech(:) = DZERO
    rech_cflag(:) = 0

    ! set recharge to cell
    if (rech_flag == 0) then  !classification
      call set_class2surf(totaln_rech, rech_class_name, rech_value, cell_rech, rech_cflag)
    else if (rech_flag == 1) then  !point
      call set_point2surf(totaln_rech, rech_i, rech_j, rech_value, cell_rech, rech_cflag)
    else if (rech_flag == 2) then
      call read_2darray(17, runi_conv, cell_rech)
      rech_cflag(1:ncals) = 1
    end if

    ! allocate
    allocate(cell_well(ncalc))
    allocate(well_cflag(ncalc))
    cell_well(:) = DZERO
    well_cflag(:) = 0

    ! set well to cell
    if (well_flag == 0) then  !classification
      call set_class2cell2(totaln_well, well_class_name, well_value, cell_well, well_cflag)
    else if (well_flag == 1) then  !point
      call set_point2cell(totaln_well, well_i, well_j, well_k, well_value, cell_well, well_cflag)
    end if

    ! allocate
    allocate(cell_bound(ncalc))
    allocate(bound_cflag(ncalc))
    cell_bound(:) = DZERO
    bound_cflag(:) = 0

    ! set boundary to cell
    if (bound_flag == 0) then  !classification
      call set_class2cell2(totaln_bound, bound_class_name, bound_value, cell_bound, bound_cflag)
    else if (bound_flag == 1) then  !point
      call set_point2cell(totaln_bound, bound_i, bound_j, bound_k, bound_value, cell_bound, bound_cflag)
    end if

    ! allocate
    allocate(cell_surf(ncals))
    allocate(surf_cflag(ncals))
    cell_surf(:) = DZERO
    surf_cflag(:) = 0

    ! set surface water boundary to cell
    if (surf_flag == 0) then  !classification
      call set_class2surf(totaln_surf, surf_class_name, surf_value, cell_surf, surf_cflag)
    else if (surf_flag == 1) then  !point
      call set_point2surf(totaln_surf, surf_i, surf_j, surf_value, cell_surf, surf_cflag)
    end if

    ! allocate
    allocate(seep_cflag(ncals))
    seep_cflag(:) = 0

    ! set seepage setting  to cell
    if (seep_flag == 0) then  !classification
      call set_class2seep(totaln_seep, seep_class_name, seep_cflag)
    else if (seep_flag == 1) then  !point
      call set_point2seep(totaln_seep, seep_i, seep_j, seep_cflag)
    end if

    ! if multi grid method is used
    if (solver_flag >= 5) then
      ! -- Set multigrid calculation (mgcalc)
        call set_mgcalc()
    end if

    ! -- Set connectivity (connect)
      call set_connect(bound_cflag, ng, nghost)

    if (solver_flag >= 5) then
      ! -- Set restriction information (resinfo)
        call set_resinfo(bound_cflag)
      ! -- Set prolongation information (proinfo)
        call set_proinfo(bound_cflag)
    end if

    ! -- Check calculation condition (calc)
      call check_calc()

    ! allocate
    allocate(cell_head(ncalc), ghost_head(nghost), surf_head(ncals))
    cell_head(:) = DZERO ; ghost_head(:) = DZERO ; surf_head(:) = DZERO

    do i = 1, ncalc
      s = calc2cals(i)
      if (s /= 0) then
        if (surf_cflag(s) == 1) then
          surf_head(s) = cell_surf(s)
        else
          surf_head(s) = cell_init(i)
        end if
      end if
      if (bound_cflag(i) == 1) then
        cell_head(i) = cell_bound(i)
      else if (bound_cflag(i) == 2) then
        do g = g_index(i-1)+1, g_index(i)
          ghost_head(g) = ghost_bound(i)
        end do
        cell_head(i) = cell_init(i)
      else
        cell_head(i) = cell_init(i)
      end if
    end do

    deallocate(cell_init)

    return

  end subroutine prepare_calc

  subroutine set_mgcalc()
  !*****************************************************************************
  ! set_mgcalc -- Set multigrid calculation
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Set mg volume (mgvol)
      call set_mgvol(vol_c)

    ! -- Set mg condition volume average (mgvolave)
      call set_mgvolave(cell_reta)
      call set_mgvolave(cell_retn)
      call set_mgvolave(cell_res)

    ! -- Set mg condition volume average (mgvolave)
      call set_mgvolave(cell_ksx)
      call set_mgvolave(cell_ksy)
      call set_mgvolave(cell_ksz)
      call set_mgvolave(cell_ss)
      call set_mgvolave(cell_sat)

    ! -- Set mg condition volume average (mgvolave)
      call set_mgvolave(cell_init)

    ! -- Set mg area (mgare)
      call set_mgare(area)

    if (rech_flag >= 0) then
      ! -- Set mg condition area average (mgareave)
        call set_mgareave(cell_rech, rech_cflag)
    end if

    if (well_flag >= 0) then
      ! -- Set mg condition summation (mgsum)
        call set_mgsum(cell_well, well_cflag)
    end if

    if (bound_flag >= 0) then
      allocate(ghost_bound(ncalc))
      ghost_bound(:) = DZERO
      ! -- Set mg boundary condition (mgbound)
        call set_mgbound(cell_init, cell_bound, bound_cflag, ghost_bound, ng)
    end if

    if (surf_flag >= 0) then
      ! -- Set mg condition area average (mgareave)
        call set_mgareave(cell_surf, surf_cflag)
    end if

    if (seep_flag >= 0) then
      ! -- Set mg condition area average (mgareave)
        call set_mgseep(seep_cflag)
    end if

    return

  end subroutine set_mgcalc

  subroutine check_calc()
  !*****************************************************************************
  ! check_calc -- Set multigrid calculation
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    check_flag = 0

    ! -- Check cell value (cellvalue)
      call check_cellvalue(cell_reta, check_flag)
      call check_cellvalue(reten_n, check_flag)
      call check_cellvalue(resi_theta, check_flag)
    if (check_flag > 0) then
      write(*,'(a)') "There is a cell which has no retention value!!"
    else
      check_flag = 0
    end if

    ! -- Check cell value (cellvalue)
      call check_cellvalue(cell_ksx, check_flag)
      call check_cellvalue(cell_ksy, check_flag)
      call check_cellvalue(cell_ksz, check_flag)
      call check_cellvalue(cell_ss, check_flag)
      call check_cellvalue(cell_sat, check_flag)
    if (check_flag > 0) then
      write(*,'(a)') "There is a cell which has no parameter value!!"
      stop
    end if

    ! -- Check cell value (cellvalue)
      call check_cellvalue(cell_init, check_flag)
    if (check_flag > 0) then
      write(*,'(a)') "There is a cell which has no initial value!!"
      stop
    end if

    return

  end subroutine check_calc

  subroutine check_cellvalue(cellv, cflag)
  !*****************************************************************************
  ! check_cellvalue -- Check cell value
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: cellv(ncalc)
    integer(I4), intent(inout) :: cflag
    ! -- local
    !---------------------------------------------------------------------------
    do i = 1, ncalc
      if (cellv(i) == DNOVAL) then
        cflag = cflag + 1
        exit
      end if
    end do

    return

  end subroutine check_cellvalue

end module set_calculation
