module set_classification
  ! -- modules
  use kind_module, only: I4, DP
  use read_inputfile, only: nx, ny, nz, nxyz, nlevel, solver_flag, class_num,&
                            totaln_class, class_name, class_i, class_j, class_k,&
                            calcout_flag, calcout_name, outcalgrd_flag

  implicit none
  private
  public :: set_class, get_calgrd
  integer(I4), allocatable, public :: nxgrid(:), nygrid(:), nzgrid(:)
  integer(I4), public :: ncell, nsurf, maxghost
  integer(I4), allocatable, public :: c_end(:), s_end(:)
  integer(I4), allocatable, public :: class_flag(:,:)
  integer(I4), public :: ncalc, ncals, nunknow
  integer(I4), allocatable, public :: calc_end(:), cals_end(:), out_flag(:)
  integer(I4), allocatable, public :: calc2cell(:), cell2calc(:), calc2cals(:)
  integer(I4), allocatable, public :: cals2calc(:), cals2cell(:), cell2cals(:)
  integer(I4), allocatable, public :: cals2surf(:), surf2cals(:)
  integer(I4), allocatable, public :: cell2surf(:), surf2cell(:)

  ! -- local
  integer(I4) :: i, j, k, ii, jj, kk, n, c_num, i_num, j_num, k_num

  contains

  subroutine set_class()
  !*****************************************************************************
  ! set_class -- Set classification & calculation cell
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: ndivide
    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(nxgrid(nlevel), nygrid(nlevel), nzgrid(nlevel))
    allocate(c_end(0:nlevel), s_end(0:nlevel))

    nxgrid(:) = 0 ; nygrid(:) = 0 ; nzgrid(:) = 0
    c_end(:) = 0 ; s_end(:) = 0

    if (solver_flag >= 5) then ! if multi grid method is used
      ndivide = 2**(nlevel-1)
      if (mod(nx,ndivide) /= 0 .or. mod(ny,ndivide) /= 0 .or. mod(nz,ndivide) /= 0) then
        write(*,'(a)') "Error wrong setting. check the number of grid!!"
        stop
      else
        ncell = 0 ; nsurf = 0 ; maxghost = 0
        do n = 1, nlevel
          nxgrid(n) = nx/ndivide*(2**(nlevel-n))
          nygrid(n) = ny/ndivide*(2**(nlevel-n))
          nzgrid(n) = nz/ndivide*(2**(nlevel-n))
          ncell = ncell + nxgrid(n)*nygrid(n)*nzgrid(n)
          nsurf = nsurf + nxgrid(n)*nygrid(n)
          c_end(n) = ncell
          s_end(n) = nsurf
          if (n > 1) then
            maxghost = maxghost + (nxgrid(n)+nygrid(n))*nzgrid(n)*2
          end if
        end do
      end if
    else
      nlevel = 1
      nxgrid(1) = nx
      nygrid(1) = ny
      nzgrid(1) = nz
      ncell = nx*ny*nz
      nsurf = nx*ny
      c_end(1) = ncell
      s_end(1) = nsurf
    end if

    ! -- Make cell classification (cellclass)
      call make_cellclass()

    ! -- Make calculation cell (calccell)
      call make_calccell()

    if (outcalgrd_flag > 0) then
      write(200,'(a)') "Calclation Number I Number J Number K Number"
      do i = 1, calc_end(1)
      ! -- Get relationship of calculation & grid number (calgrd)
        call get_calgrd(i, i_num, j_num, k_num)
        write(200,'(i17,3i9)') i, i_num, j_num, k_num
      end do
      close(200)
    end if

    return

  end subroutine set_class

  subroutine make_cellclass()
  !*****************************************************************************
  ! make_cellclass -- Make cell classification
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(class_flag(totaln_class,nxyz))
    class_flag(:,:) = 0

    do i = 1, totaln_class
      do j = 1, class_num(i)
        if (class_i(i,j) == -1 .and. class_j(i,j) == -1 .and. class_k(i,j) == -1) then !all cells
          class_flag(i,1:nxyz) = 1
        else if (class_i(i,j) == -1 .and. class_j(i,j) == -1) then !i,j cells
          do jj = 1, ny
            do ii = 1, nx
              c_num = nx*ny*(class_k(i,j)-1) + nx*(jj-1) + ii
              class_flag(i,c_num) = 1
            end do
          end do
        else if (class_i(i,j) == -1 .and. class_k(i,j) == -1) then !i,k cells
          do kk = 1, nz
            do ii = 1, nx
              c_num = nx*ny*(kk-1) + nx*(class_j(i,j)-1) + ii
              class_flag(i,c_num) = 1
            end do
          end do
        else if (class_j(i,j) == -1 .and. class_k(i,j) == -1) then !j,k cells
          do kk = 1, nz
            do jj = 1, ny
              c_num = nx*ny*(kk-1) + nx*(jj-1) + class_i(i,j)
              class_flag(i,c_num) = 1
            end do
          end do
        else if (class_i(i,j) == -1) then !only i cell
          do ii = 1, nx
            c_num = nx*ny*(class_k(i,j)-1) + nx*(class_j(i,j)-1) + ii
            class_flag(i,c_num) = 1
          end do
        else if (class_j(i,j) == -1) then !only j cell
          do jj = 1, ny
            c_num = nx*ny*(class_k(i,j)-1) + nx*(jj-1) + class_i(i,j)
            class_flag(i,c_num) = 1
          end do
        else if (class_k(i,j) == -1) then !only k cell
          do kk = 1, nz
            c_num = nx*ny*(kk-1) + nx*(class_j(i,j)-1) + class_i(i,j)
            class_flag(i,c_num) = 1
          end do
        else !others
          c_num = nx*ny*(class_k(i,j)-1) + nx*(class_j(i,j)-1) + class_i(i,j)
          class_flag(i,c_num) = 1
        end if
      end do
    end do

    return

  end subroutine make_cellclass

  subroutine make_calccell()
  !*****************************************************************************
  ! make_calccell -- Make calculation cell
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: calcout_class, nxygrid, flag_sum, s_num
    integer(I4) :: x1, x2, y1, y2, z1, z2, z3, z4
    !---------------------------------------------------------------------------
    if (calcout_flag == 0) then
      ncalc = ncell ; ncals = nsurf
    else if (calcout_flag == 1) then
      ncalc = 0 ; ncals = 0
      do i = 1, totaln_class
        if (trim(adjustl(calcout_name)) == trim(adjustl(class_name(i)))) then
          ncalc = ncell - sum(class_flag(i,:))
          calcout_class = i
          do s_num = s_end(0)+1, s_end(1)
            if (class_flag(i,s_num) == 1) then
              ncals = ncals + 1
            end if
          end do
          ncals = nsurf - ncals
        end if
      end do
    end if

    ! -- allocate
    allocate(calc_end(0:nlevel), cals_end(0:nlevel), out_flag(ncell))
    allocate(calc2cell(ncalc), cell2calc(ncell), calc2cals(ncalc))
    allocate(cals2calc(ncals), cals2cell(ncals), cell2cals(ncell))
    allocate(cals2surf(ncals), surf2cals(nsurf), cell2surf(ncell))
    allocate(surf2cell(nsurf))

    calc_end(:) = 0 ; cals_end(:) = 0 ; out_flag(:) = 0
    calc2cell(:) = 0 ; cell2calc(:) = 0 ; calc2cals(:) = 0 ; cals2calc(:) = 0
    cals2cell(:) = 0 ; cell2cals(:) = 0 ; cals2surf(:) = 0 ; surf2cals(:) = 0
    cell2surf(:) = 0 ; surf2cell(:) = 0

    ncalc = 0
    ncals = 0

    if (calcout_flag == 0) then
      do n = 1, nlevel
        nxygrid = nxgrid(n)*nygrid(n)
        do i = c_end(n-1)+1, c_end(n)
          ncalc = ncalc + 1
          calc2cell(ncalc) = i
          cell2calc(i) = ncalc
          if (i <= c_end(n-1)+nxygrid) then
            ncals = ncals + 1
            s_num = s_end(n-1) + i - c_end(n-1)
            cals2cell(ncals) = i
            cell2cals(i) = ncals
            cals2surf(ncals) = s_num
            surf2cals(s_num) = ncals
            cals2calc(ncals) = ncalc
            calc2cals(ncalc) = ncals
            surf2cell(s_num) = i
            cell2surf(i) = s_num
          end if
        end do
      calc_end(n) = ncalc
      cals_end(n) = ncals
      if (n == 1) then
        nunknow = ncalc
      end if
      end do
    else if (calcout_flag == 1) then
      do n = 1, nlevel
        nxygrid = nxgrid(n)*nygrid(n)
        if (n == 1) then
          do j = 1, nxyz
            if (class_flag(calcout_class,j) /= 0) then
              cell2calc(j) = 0
              out_flag(j) = 1
              if (j <= nx*ny) then
                s_num = s_end(n-1) + j - c_end(n-1)
                surf2cell(j) = j
                cell2surf(j) = j
              end if
            else
              ncalc = ncalc + 1
              calc2cell(ncalc) = j
              cell2calc(j) = ncalc
              out_flag(j) = 0
              if (j <= nx*ny) then
                ncals = ncals + 1
                s_num = s_end(n-1) + j - c_end(n-1)
                cals2cell(ncals) = j
                cell2cals(j) = ncals
                cals2surf(ncals) = s_num
                surf2cals(s_num) = ncals
                cals2calc(ncals) = ncalc
                calc2cals(ncalc) = ncals
                surf2cell(s_num) = j
                cell2surf(j) = s_num
              end if
            end if
          end do
        calc_end(n) = ncalc
        cals_end(n) = ncals
        nunknow = ncalc
        else
          do k = 1, nzgrid(n)
            do j = 1, nygrid(n)
              do i = 1, nxgrid(n)
                c_num = c_end(n-1) + nxygrid*(k-1) + nxgrid(n)*(j-1) + i
                x1 = c_end(n-2) + (nxgrid(n-1)*nygrid(n-1)*(k-1) + nxgrid(n-1)*(j-1) + i)*2 - 1
                x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
                z1 = x1 + nxgrid(n-1)*nygrid(n-1) ; z2 = z1 + 1
                z3 = z1 + nxgrid(n-1) ; z4 = z3 + 1
                flag_sum = out_flag(x1) + out_flag(x2) + out_flag(y1)&
                         + out_flag(y2) + out_flag(z1) + out_flag(z2)&
                         + out_flag(z3) + out_flag(z4)
                if (flag_sum == 8) then
                  cell2calc(c_num) = 0
                  out_flag(c_num) = 1
                  if (c_num <= c_end(n-1)+nxygrid) then
                    s_num = s_end(n-1) + c_num - c_end(n-1)
                    surf2cell(s_num) = c_num
                    cell2surf(c_num) = s_num
                  end if
                else
                  ncalc = ncalc + 1
                  calc2cell(ncalc) = c_num
                  cell2calc(c_num) = ncalc
                  out_flag(c_num) = 0
                  if (c_num <= c_end(n-1)+nxygrid) then
                    ncals = ncals + 1
                    s_num = s_end(n-1) + c_num - c_end(n-1)
                    cals2cell(ncals) = c_num
                    cell2cals(c_num) = ncals
                    cals2surf(ncals) = s_num
                    surf2cals(s_num) = ncals
                    cals2calc(ncals) = ncalc
                    calc2cals(ncalc) = ncals
                    surf2cell(s_num) = c_num
                    cell2surf(c_num) = s_num
                  end if
                end if
              end do
            end do
          end do
          calc_end(n) = ncalc
          cals_end(n) = ncals
        end if
      end do
    end if

    return

  end subroutine make_calccell

  subroutine get_calgrd(cal_num, x_num, y_num, z_num)
  !*****************************************************************************
  ! get_calgrd -- Get relationship of calculation & grid number
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: cal_num
    integer(I4), intent(out) :: x_num, y_num, z_num
    ! -- local
    integer(I4) :: c_num, xy_num
    !---------------------------------------------------------------------------
    c_num = calc2cell(cal_num)
    z_num = (c_num-1)/(nx*ny) + 1
    xy_num = c_num - nx*ny*(z_num-1)
    y_num = (xy_num-1)/nx + 1
    x_num = xy_num - (y_num-1)*nx

    return

  end subroutine get_calgrd

end module set_classification
