module set_condition
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, FACE, FILELENGTH, DNOVAL
  use read_inputfile, only: nx, ny, nz, nxyz, nlevel, totaln_class, class_name,&
                            init_depth, solver_flag
  use set_classification, only: ncalc, ncals, nunknow, out_flag, c_end,&
                                nxgrid, nygrid, nzgrid, cell2calc, surf2cals,&
                                calc2cell, class_flag
  use make_cell, only: surfelev

  implicit none
  private
  public :: set_class2cell1, set_class2cell2, set_class2surf, set_class2seep
  public :: set_point2cell, set_point2surf, set_point2seep
  public :: read_3darray, read_2darray
  public :: set_glevel, set_connect, set_class2out, set_point2out
  integer(I4), allocatable, public :: off_row(:), conn_dir(:), off_index(:)
  integer(I4), allocatable, public :: g_index(:), g_dir(:)

  ! -- local
  integer(I4) :: i, j, k, ii, jj, kk, n
  integer(I4) :: c_num, c, s_num, s

  contains

  subroutine set_class2cell1(targetn, target_name, target_value, cell_value)
  !**********************************************************************
  ! set_class2cell1 -- Set cell from classification
  !**********************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    character(len=*), intent(in) :: target_name(:)
    real(DP), intent(in) :: target_value(:)
    real(DP), intent(out) :: cell_value(ncalc)
    ! -- local

    !---------------------------------------------------------------------------
    cell_value(:) = DNOVAL

    do i = 1, targetn
      do j = 1, totaln_class
        if (target_name(i) == class_name(j)) then
          do k = 1, nxyz
            if (class_flag(j,k) == 1 .and. out_flag(k) == 0) then
              c = cell2calc(k)
              cell_value(c) = target_value(i)
            end if
          end do
        end if
      end do
    end do

    return

  end subroutine set_class2cell1

  subroutine set_class2cell2(targetn, target_name, target_value, cell_value, target_flag)
  !**********************************************************************
  ! set_class2cell2 -- Set cell from classification
  !**********************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    character(len=*), intent(in) :: target_name(:)
    real(DP), intent(in) :: target_value(:)
    real(DP), intent(out) :: cell_value(ncalc)
    integer(I4), intent(out) :: target_flag(ncalc)
    ! -- local

    !---------------------------------------------------------------------------
    cell_value(:) = DZERO ; target_flag(:) = 0

    do i = 1, targetn
      do j = 1, totaln_class
        if (target_name(i) == class_name(j)) then
          do k = 1, nxyz
            if (class_flag(j,k) == 1 .and. out_flag(k) == 0) then
              c = cell2calc(k)
              cell_value(c) = target_value(i)
              target_flag(c) = 1
            end if
          end do
        end if
      end do
    end do

    return

  end subroutine set_class2cell2

  subroutine set_class2surf(targetn, target_name, target_value, cell_value, target_flag)
  !**********************************************************************
  ! set_class2surf -- Set surface from classification
  !**********************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    character(len=*), intent(in) :: target_name(:)
    real(DP), intent(in) :: target_value(:)
    real(DP), intent(out) :: cell_value(ncals)
    integer(I4), intent(out) :: target_flag(ncals)
    ! -- local

    !---------------------------------------------------------------------------
    cell_value(:) = DZERO ; target_flag(:) = 0

    do i = 1, targetn
      do j = 1, totaln_class
        if (target_name(i) == class_name(j)) then
          do k = 1, nx*ny
            if (class_flag(j,k) == 1 .and. out_flag(k) == 0) then
              s = surf2cals(k)
              cell_value(s) = target_value(i)
              target_flag(s) = 1
            end if
          end do
        end if
      end do
    end do

    return

  end subroutine set_class2surf

  subroutine set_class2seep(targetn, target_name, target_flag)
  !**********************************************************************
  ! set_class2seep -- Set seepage from classification
  !**********************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    character(len=*), intent(in) :: target_name(:)
    integer(I4), intent(out) :: target_flag(ncals)
    ! -- local

    !---------------------------------------------------------------------------
    target_flag(:) = 0

    do i = 1, targetn
      do j = 1, totaln_class
        if (target_name(i) == class_name(j)) then
          do k = 1, nx*ny
            if (class_flag(j,k) == 1 .and. out_flag(k) == 0) then
              s = surf2cals(k)
              target_flag(s) = 1
            end if
          end do
        end if
      end do
    end do

    return

  end subroutine set_class2seep

  subroutine set_point2cell(targetn, p_i, p_j, p_k, target_value, cell_value, target_flag)
  !*****************************************************************************
  ! set_point2cell -- Set cell from point
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    integer(I4), intent(in) :: p_i(:), p_j(:), p_k(:)
    real(DP), intent(in) :: target_value(:)
    real(DP), intent(out) :: cell_value(ncalc)
    integer(I4), intent(out) :: target_flag(ncalc)
    ! -- local
    integer(I4) :: ista, iend, jsta, jend, ksta, kend
    !---------------------------------------------------------------------------
    cell_value(:) = DZERO ; target_flag(:) = 0

    do i = 1, targetn
      if (p_i(i) == -1 .and. p_j(i) == -1 .and. p_k(i) == -1) then !all cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny ; ksta = 1 ; kend = nz
      else if (p_i(i) == -1 .and. p_j(i) == -1) then !i,j cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_i(i) == -1 .and. p_k(i) == -1) then !i,k cells
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i) ; ksta = 1 ; kend = nz
      else if (p_j(i) == -1 .and. p_k(i) == -1) then !j,k cells
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny ; ksta = 1 ; kend = nz
      else if (p_i(i) == -1) then !only i cell
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i) ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_j(i) == -1) then !only j cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_k(i) == -1) then !only k cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i) ; ksta = 1 ; kend = nz
      else !others
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i) ; ksta = p_k(i) ; kend = p_k(i)
      end if

      do ii = ista, iend
        do jj = jsta, jend
          do kk = ksta, kend
            c_num = nx*ny*(kk-1) + nx*(jj-1) + ii
            if (out_flag(c_num) == 0) then
              c = cell2calc(c_num)
              cell_value(c) = target_value(i)
              target_flag(c) = 1
            end if
          end do
        end do
      end do

    end do

    return

  end subroutine set_point2cell

  subroutine set_point2surf(targetn, p_i, p_j, target_value, cell_value, target_flag)
  !*****************************************************************************
  ! set_point2surf -- Set surface from point
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    integer(I4), intent(in) :: p_i(:), p_j(:)
    real(DP), intent(in) :: target_value(:)
    real(DP), intent(out) :: cell_value(ncals)
    integer(I4), intent(out) :: target_flag(ncals)
    ! -- local
    integer(I4) :: ista, iend, jsta, jend
    !---------------------------------------------------------------------------
    cell_value(:) = DZERO ; target_flag(:) = 0

    do i = 1, targetn
      if (p_i(i) == -1 .and. p_j(i) == -1) then !all i,j cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny
      else if (p_i(i) == -1) then !only i cell
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i)
      else if (p_j(i) == -1) then !only j cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny
      else !others
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i)
      end if

      do ii = ista, iend
        do jj = jsta, jend
          s_num = nx*(jj-1) + ii
          if (out_flag(s_num) == 0) then
            s = surf2cals(s_num)
            cell_value(s) = target_value(i)
            target_flag(s) = 1
          end if
        end do
      end do

    end do

    return

  end subroutine set_point2surf

  subroutine set_point2seep(targetn, p_i, p_j, target_flag)
  !*****************************************************************************
  ! set_point2seep -- Set seepage from point
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    integer(I4), intent(in) :: p_i(:), p_j(:)
    integer(I4), intent(out) :: target_flag(ncals)
    ! -- local
    integer(I4) :: ista, iend, jsta, jend
    !---------------------------------------------------------------------------
    target_flag(:) = 0

    do i = 1, targetn
      if (p_i(i) == -1 .and. p_j(i) == -1) then !all i,j cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny
      else if (p_i(i) == -1) then !only i cell
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i)
      else if (p_j(i) == -1) then !only j cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny
      else !others
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i)
      end if

      do ii = ista, iend
        do jj = jsta, jend
          s_num = nx*(jj-1) + ii
          if (out_flag(s_num) == 0) then
            s = surf2cals(s_num)
            target_flag(s) = 1
          end if
        end do
      end do

    end do

    return

  end subroutine set_point2seep

  subroutine read_3darray(fnum, array_value)
  !*****************************************************************************
  ! read_3darray -- Read value from 3d array file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: fnum
    real(DP), intent(out) :: array_value(ncalc)
    ! -- local
    integer(I4) :: a_num
    real(DP), allocatable :: array_in(:)
    !---------------------------------------------------------------------------
    ! allocate
    allocate(array_in(nxyz))
    array_in(:) = DZERO

    do k = 1, nz
      do j = ny, 1, -1
        a_num = nx*ny*(k-1) + nx*(j-1)
        read(fnum, *) (array_in(a_num+i), i = 1, nx)
        do i = 1, nx
          c_num = a_num + i
          if (out_flag(c_num) == 0) then
            c = cell2calc(c_num)
            array_value(c) = array_in(a_num+i)
          end if
        end do
      end do
    end do

    deallocate(array_in)

    return

  end subroutine read_3darray

  subroutine read_2darray(fnum, inunit, array_value)
  !*****************************************************************************
  ! read_2darray -- Read value from 2d array file
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: fnum
    real(DP), intent(in) :: inunit
    real(DP), intent(out) :: array_value(ncals)
    ! -- local
    integer(I4) :: a_num
    real(DP), allocatable :: array_in(:)
    !---------------------------------------------------------------------------
    ! allocate
    allocate(array_in(nx*ny))
    array_in(:) = DZERO

    do j = ny, 1, -1
      a_num = nx*(j-1)
      read(fnum, *) (array_in(a_num+i), i = 1, nx)
      do i = 1, nx
        s_num = a_num + i
        if (out_flag(s_num) == 0) then
          s = surf2cals(s_num)
          array_value(s) = array_in(s_num)*inunit
        end if
      end do
    end do

    deallocate(array_in)

    return

  end subroutine read_2darray

  subroutine set_glevel(cell_value)
  !*****************************************************************************
  ! set_glevel -- Set initial waterlevel from surface elevation
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(out) :: cell_value(ncalc)
    ! -- local
    !---------------------------------------------------------------------------
    do k = 1, nz
      do j = 1, nx*ny
        c_num = nx*ny*(k-1) + j
        if (out_flag(c_num) == 0) then
          c = cell2calc(c_num)
          cell_value(c) = surfelev(j) - init_depth
        end if
      end do
    end do

    return

  end subroutine set_glevel

  subroutine set_connect(bflag, gn, ghostcount)
  !*****************************************************************************
  ! set_connect -- Set connectivity
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: bflag(ncalc)
    integer(I4), intent(in) :: gn
    integer(I4), intent(out) :: ghostcount
    ! -- local
    integer(I4) :: tconn_num, offdaig, offcount, surfcount
    integer(I4) :: calc_num, off_num
    !---------------------------------------------------------------------------
    allocate(off_row(ncalc*FACE), conn_dir(ncalc*FACE), off_index(0:ncalc))
    off_row(:) = 0 ; conn_dir(:) = 0 ; off_index(:) = 0

    if (solver_flag >= 5) then
      allocate(g_index(0:ncalc), g_dir(gn*4))
      g_index(:) = 0 ; g_dir(:) = 0
    end if

    tconn_num = 0

    do n = 1, nlevel
      if (n == 1) then
        offcount = 0
!        surfcount = 0
        ghostcount = 0
      else
        offcount = offcount + nxgrid(n-1)*nygrid(n-1)*nzgrid(n-1)
      end if
      do k = 1, nzgrid(n)
        do j = 1, nygrid(n)
          do i = 1, nxgrid(n)
            c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + i
            calc_num = cell2calc(c_num)
            if (out_flag(c_num) == 1) then
              cycle
            else if (bflag(calc_num) == 1) then
              off_index(calc_num) = tconn_num
              if  (solver_flag >= 5) then
                g_index(calc_num) = ghostcount
              end if
!              if (k == 1) then
!                surfcount = surfcount + 1
!              end if
              cycle
            else if (out_flag(c_num) == 0) then
              ! up direction
              if (k > 1) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1-1) + nxgrid(n)*(j-1) + i
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 1
                end if
!              else
!                surfcount = surfcount + 1
!                tconn_num = tconn_num + 1
!                off_row(tconn_num) = surfcount
!                conn_dir(tconn_num) = 1
              end if

              ! south direction
              if (j > 1) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1-1) + i
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 1 .and. bflag(calc_num) == 2) then
                  ghostcount = ghostcount + 1
                  g_dir(ghostcount) = 2
!                  calc2ghost(calc_num) = ghostcount
                else if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 2
                end if
!                if (j == nygrid(n) .and. n > 1) then
!                  if (bflag(off_num) /= 1 .and. bflag(calc_num) == 2) then
!                    ghostcount = ghostcount + 1
!                    g_dir(ghostcount) = 5
!                  end if
!                end if
              else if (bflag(calc_num) == 2) then
                ghostcount = ghostcount + 1
                g_dir(ghostcount) = 2
!                calc2ghost(calc_num) = ghostcount
              end if

              ! west direction
              if (i > 1) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + (i-1)
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 1 .and. bflag(calc_num) == 2) then
                  ghostcount = ghostcount + 1
                  g_dir(ghostcount) = 3
!                  calc2ghost(calc_num) = ghostcount
                else if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 3
                end if
!                if (i == nxgrid(n) .and. n > 1) then
!                  if (bflag(off_num) /= 1 .and. bflag(calc_num) == 2) then
!                    ghostcount = ghostcount + 1
!                    g_dir(ghostcount) = 4
!                  end if
!                end if
              else if (bflag(calc_num) == 2) then
                ghostcount = ghostcount + 1
                g_dir(ghostcount) = 3
!                calc2ghost(calc_num) = ghostcount
              end if

              ! east direction
              if (i < nxgrid(n)) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + (i+1)
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 1 .and. bflag(calc_num) == 2) then
                  ghostcount = ghostcount + 1
                  g_dir(ghostcount) = 4
!                  calc2ghost(calc_num) = ghostcount
                else if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 4
                end if
!                if (i == 1 .and. n > 1) then
!                  if (bflag(off_num) /= 1 .and. bflag(calc_num) == 2) then
!                    ghostcount = ghostcount + 1
!                    g_dir(ghostcount) = 3
!                  end if
!                end if
              else if (bflag(calc_num) == 2) then
                ghostcount = ghostcount + 1
                g_dir(ghostcount) = 4
!                calc2ghost(calc_num) = ghostcount
              end if

              ! north direction
              if (j < nygrid(n)) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1+1) + i
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 1 .and. bflag(calc_num) == 2) then
                  ghostcount = ghostcount + 1
                  g_dir(ghostcount) = 5
!                  calc2ghost(calc_num) = ghostcount
                else if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 5
                end if
!                if (j == 1 .and. n > 1) then
!                  if (bflag(off_num) /= 1 .and. bflag(calc_num) == 2) then
!                    ghostcount = ghostcount + 1
!                    g_dir(ghostcount) = 2
!                  end if
!                end if
              else if (bflag(calc_num) == 2) then
                ghostcount = ghostcount + 1
                g_dir(ghostcount) = 5
!                calc2ghost(calc_num) = ghostcount
              end if

              ! down direction
              if (k < nzgrid(n)) then
                offdaig = offcount + nxgrid(n)*nygrid(n)*(k-1+1) + nxgrid(n)*(j-1) + i
                off_num = cell2calc(offdaig)
                if (out_flag(offdaig) == 0) then
                  tconn_num = tconn_num + 1
                  off_row(tconn_num) = off_num
                  conn_dir(tconn_num) = 6
                end if
              end if
            end if
            off_index(calc_num) = tconn_num
            if  (solver_flag >= 5) then
              g_index(calc_num) = ghostcount
            end if
          end do
        end do
      end do
    end do

    off_index(0) = 0
    if  (solver_flag >= 5) then
      g_index(0) = 0
    end if

    return

  end subroutine set_connect

  subroutine set_class2out(targetn, target_name, target_index, target_cnum, num)
  !*****************************************************************************
  ! set_class2out -- Set output cell from classification
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    character(len=*), dimension(targetn), intent(in) :: target_name(:)
    integer(I4), intent(out) :: target_index(targetn)
    integer(I4), intent(out) :: target_cnum(targetn*nunknow)
    integer(I4), intent(out) :: num
    ! -- local
    !---------------------------------------------------------------------------
    num = 0

    do i = 1, targetn
      do j = 1, totaln_class
        if (target_name(i) == class_name(j)) then
          do k = 1, nunknow
            c_num = calc2cell(k)
            if (class_flag(j,c_num) == 1 .and. out_flag(c_num) == 0) then
              num = num + 1
              target_cnum(num) = k
            end if
          end do
        end if
      end do
      target_index(i) = num
    end do

    return

  end subroutine set_class2out

  subroutine set_point2out(targetn, p_i, p_j, p_k, wri, wrj, wrk, target_cnum, num)
  !*****************************************************************************
  ! set_point2out -- Set output cell from point
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: targetn
    integer(I4), intent(in) :: p_i(:), p_j(:), p_k(:)
    integer(I4), intent(out) :: wri(nunknow), wrj(nunknow), wrk(nunknow)
    integer(I4), intent(out) :: target_cnum(nunknow)
    integer(I4), intent(out) :: num
    ! -- local
    integer(I4) :: ista, iend, jsta, jend, ksta, kend
    !---------------------------------------------------------------------------
    num = 0

    do i = 1, targetn
      if (p_i(i) == -1 .and. p_j(i) == -1 .and. p_k(i) == -1) then !all cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny ; ksta = 1 ; kend = nz
      else if (p_i(i) == -1 .and. p_j(i) == -1) then !i,j cells
        ista = 1 ; iend = nx ; jsta = 1 ; jend = ny ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_i(i) == -1 .and. p_k(i) == -1) then !i,k cells
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i) ; ksta = 1 ; kend = nz
      else if (p_j(i) == -1 .and. p_k(i) == -1) then !j,k cells
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny ; ksta = 1 ; kend = nz
      else if (p_i(i) == -1) then !only i cell
        ista = 1 ; iend = nx ; jsta = p_j(i) ; jend = p_j(i) ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_j(i) == -1) then !only j cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = 1 ; jend = ny ; ksta = p_k(i) ; kend = p_k(i)
      else if (p_k(i) == -1) then !only k cell
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i) ; ksta = 1 ; kend = nz
      else !others
        ista = p_i(i) ; iend = p_i(i) ; jsta = p_j(i) ; jend = p_j(i) ; ksta = p_k(i) ; kend = p_k(i)
      end if

      do kk = ksta, kend
        do jj = jsta, jend
          do ii = ista, iend
            num = num + 1
            wri(num) = ii
            wrj(num) = jj
            wrk(num) = kk
            c_num = nx*ny*(kk-1) + nx*(jj-1) + ii
            c = cell2calc(c_num)
            if (out_flag(c_num) == 0) then
              target_cnum(num) = c
            end if
          end do
        end do
      end do

    end do

    return

  end subroutine set_point2out

end module set_condition
