module set_mg_condition
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DHALF, DONE, FACE
  use read_inputfile, only: nlevel
  use set_classification, only: ncell, nsurf, ncalc, out_flag, c_end, s_end,&
                                nxgrid, nygrid, nzgrid, calc_end,&
                                cell2calc, surf2cals, surf2cell, calc2cell
  use make_cell, only: area, vol_c
  use set_condition, only: off_index, conn_dir, off_row

  implicit none
  private
  public :: set_mgvol, set_mgvolave, set_mgare, set_mgareave, set_mgsum
  public :: set_mgbound, set_mgseep, set_proinfo, set_resinfo
  integer(I4), allocatable, public :: prolo_index(:), fine2coas(:), prolo_num(:)
  integer(I4), allocatable, public :: restr_index(:), restr_num(:)
  real(DP), allocatable, public :: prolo_wei(:)
  real(DP), allocatable, public :: restr_wei(:)

  ! -- local
  integer(I4) :: i, j, k, n, c, s, t, nc
  integer(I4) :: x1, x2, y1, y2, z1, z2, z3, z4
  real(DP) :: sum_value, sum_vol, sum_area
  integer(I4), allocatable :: coa_conn(:)
  integer(I4) :: fx, fy, fz, fcal_num, tcoase_num, c_num, coa_num

  contains

  subroutine set_mgvol(t_vol)
  !****************************************************************************
  ! set_mgvol -- Set mg volume
  !****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: t_vol(:)
    ! -- local

    !--------------------------------------------------------------------------
    do n = 1, nlevel
      do c = c_end(n-1)+1, c_end(n)
        if (out_flag(c) == 1) then
          t_vol(c) = DZERO
        else if (n > 1) then
          nc = c - c_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = c_end(n-2) + nxgrid(n-1)*nygrid(n-1)*(k-1) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          z1 = x1 + nxgrid(n-1)*nygrid(n-1) ; z2 = z1 + 1
          z3 = z1 + nxgrid(n-1) ; z4 = z3 + 1
          sum_vol = t_vol(x1) + t_vol(x2) + t_vol(y1) + t_vol(y2) +&
                    t_vol(z1) + t_vol(z2) + t_vol(z3) + t_vol(z4)
          t_vol(c) = sum_vol
        end if
      end do
    end do

    return

  end subroutine set_mgvol

  subroutine set_mgvolave(cell_value)
  !*****************************************************************************
  ! set_mgvolave -- Set mg condition volume average
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: cell_value(:)
    ! -- local
    real(DP), allocatable :: t_val(:), t_vol(:)
    !---------------------------------------------------------------------------
    allocate(t_val(ncell), t_vol(ncell))
    t_val(:) = DZERO ; t_vol(:) = DZERO

    do n = 1, nlevel
      do c = c_end(n-1)+1, c_end(n)
        if (n == 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          t_val(c) = cell_value(t)
          t_vol(c) = vol_c(c)
        else if (n > 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          nc = c - c_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = c_end(n-2) + nxgrid(n-1)*nygrid(n-1)*(k-1) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          z1 = x1 + nxgrid(n-1)*nygrid(n-1) ; z2 = z1 + 1
          z3 = z1 + nxgrid(n-1) ; z4 = z3 + 1
          sum_value = t_val(x1)*t_vol(x1) + t_val(x2)*t_vol(x2) +&
                      t_val(y1)*t_vol(y1) + t_val(y2)*t_vol(y2) +&
                      t_val(z1)*t_vol(z1) + t_val(z2)*t_vol(z2) +&
                      t_val(z3)*t_vol(z3) + t_val(z4)*t_vol(z4)
          sum_vol = t_vol(x1) + t_vol(x2) + t_vol(y1) + t_vol(y2) +&
                    t_vol(z1) + t_vol(z2) + t_vol(z3) + t_vol(z4)
          cell_value(t) = sum_value/sum_vol
          t_val(c) = cell_value(t) ; t_vol(c) = sum_vol
        end if
      end do
    end do

    deallocate(t_val, t_vol)

    return

  end subroutine set_mgvolave

  subroutine set_mgare(t_are)
  !****************************************************************************
  ! set_mgare -- Set mg area
  !****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: t_are(:,:)
    ! -- local

    !--------------------------------------------------------------------------
    do n = 1, nlevel
      do s = s_end(n-1)+1, s_end(n)
        c = surf2cell(s)
        if (out_flag(c) == 1) then
          t_are(c,1) = DZERO
        else if (n > 1) then
          nc = s - s_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = s_end(n-2) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          sum_area = t_are(x1,1) + t_are(x2,1) + t_are(y1,1) + t_are(y2,1)
          t_are(c,1) = sum_area
        end if
      end do
    end do

    return

  end subroutine set_mgare

  subroutine set_mgareave(cell_value, cell_flag)
  !*****************************************************************************
  ! set_mgareave -- Set mg condition area average
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: cell_value(:)
    integer(I4), intent(inout) :: cell_flag(:)
    ! -- local
    real(DP), allocatable :: t_val(:), t_are(:)
    !---------------------------------------------------------------------------
    allocate(t_val(nsurf), t_are(nsurf))
    t_val(:) = DZERO ; t_are(:) = DZERO

    do n = 1, nlevel
      do s = s_end(n-1)+1, s_end(n)
        c = surf2cell(s)
        if (n == 1 .and. out_flag(c) == 0) then
          t = surf2cals(s)
          t_val(s) = cell_value(t)
          t_are(s) = area(c,1)
        else if (n > 1 .and. out_flag(c) == 0) then
          t = surf2cals(s)
          nc = s - s_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = s_end(n-2) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          sum_value = t_val(x1)*t_are(x1) + t_val(x2)*t_are(x2) +&
                      t_val(y1)*t_are(y1) + t_val(y2)*t_are(y2)
          sum_area = t_are(x1) + t_are(x2) + t_are(y1) + t_are(y2)
          cell_value(t) = sum_value/sum_area
          t_val(s) = cell_value(t) ; t_are(s) = sum_area
          if (sum_value /= DZERO) then
            cell_flag(t) = 1
          end if
        end if
      end do
    end do

    deallocate(t_val, t_are)

    return

  end subroutine set_mgareave

  subroutine set_mgsum(cell_value, cell_flag)
  !*****************************************************************************
  ! set_mgsum -- Set mg condition summation
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: cell_value(:)
    integer(I4), intent(inout) :: cell_flag(:)
    ! -- local
    real(DP), allocatable :: t_val(:)
    !---------------------------------------------------------------------------
    allocate(t_val(ncell))
    t_val(:) = DZERO

    do n = 1, nlevel
      do c = c_end(n-1)+1, c_end(n)
        if (n == 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          t_val(c) = cell_value(t)
        else if (n > 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          nc = c - c_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = c_end(n-2) + nxgrid(n-1)*nygrid(n-1)*(k-1) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          z1 = x1 + nxgrid(n-1)*nygrid(n-1) ; z2 = z1 + 1
          z3 = z1 + nxgrid(n-1) ; z4 = z3 + 1
          sum_value = t_val(x1) + t_val(x2) + t_val(y1) + t_val(y2) +&
                      t_val(z1) + t_val(z2) + t_val(z3) + t_val(z4)
          cell_value(t) = sum_value
          t_val(c) = sum_value
          if (sum_value /= DZERO) then
            cell_flag(t) = 1
          end if
        end if
      end do
    end do

    deallocate(t_val)

    return

  end subroutine set_mgsum

  subroutine set_mgbound(cell_ivalue, cell_bvalue, cell_flag, ghost_value, tghost_num)
  !*****************************************************************************
  ! set_mgbound -- Set mg boundary condition
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: cell_ivalue(:), cell_bvalue(:)
    integer(I4), intent(inout) :: cell_flag(:)
    real(DP), intent(out) :: ghost_value(ncalc)
    integer(I4), intent(out) :: tghost_num
    ! -- local
    real(DP) :: sum_bvalue, sum_bvol
    real(DP), allocatable :: t_val(:), t_bval(:), t_vol(:), t_bvol(:)
    !---------------------------------------------------------------------------
    allocate(t_val(ncell), t_bval(ncell), t_vol(ncell), t_bvol(ncell))
    t_val(:) = DZERO ; t_bval(:) = DZERO ; t_vol(:) = DZERO ; t_bvol(:) = DZERO

    tghost_num = 0

    do n = 1, nlevel
      do c = c_end(n-1)+1, c_end(n)
        if (n == 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          t_bval(c) = cell_bvalue(t)
          t_vol(c) = vol_c(c)
          t_bvol(c) = cell_flag(t)*vol_c(c)
          if (cell_flag(t) == 1) then
            t_val(c) = cell_bvalue(t)
          else
            t_val(c) = cell_ivalue(t)
          end if
        else if (n > 1 .and. out_flag(c) == 0) then
          t = cell2calc(c)
          nc = c - c_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = c_end(n-2) + nxgrid(n-1)*nygrid(n-1)*(k-1) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          z1 = x1 + nxgrid(n-1)*nygrid(n-1) ; z2 = z1 + 1
          z3 = z1 + nxgrid(n-1) ; z4 = z3 + 1
          sum_bvalue = t_bval(x1)*t_bvol(x1) + t_bval(x2)*t_bvol(x2) +&
                       t_bval(y1)*t_bvol(y1) + t_bval(y2)*t_bvol(y2) +&
                       t_bval(z1)*t_bvol(z1) + t_bval(z2)*t_bvol(z2) +&
                       t_bval(z3)*t_bvol(z3) + t_bval(z4)*t_bvol(z4)
          sum_value = t_val(x1)*t_vol(x1) +t_val(x2)*t_vol(x2) +&
                      t_val(y1)*t_vol(y1) +t_val(y2)*t_vol(y2) +&
                      t_val(z1)*t_vol(z1) +t_val(z2)*t_vol(z2) +&
                      t_val(z3)*t_vol(z3) +t_val(z4)*t_vol(z4)
          sum_vol = t_vol(x1) + t_vol(x2) + t_vol(y1) + t_vol(y2) +&
                    t_vol(z1) + t_vol(z2) + t_vol(z3) + t_vol(z4)
          sum_bvol = t_bvol(x1) + t_bvol(x2) + t_bvol(y1) + t_bvol(y2) +&
                     t_bvol(z1) + t_bvol(z2) + t_bvol(z3) + t_bvol(z4)
          if (sum_bvol > sum_vol*DHALF) then
            cell_flag(t) = 1
            cell_bvalue(t) = sum_bvalue/sum_bvol
            t_bval(c) = cell_bvalue(t)
            cell_ivalue(t) = cell_bvalue(t) ; t_val(c) = cell_bvalue(t)
          else if (sum_bvol > DZERO) then
            tghost_num = tghost_num + 1
            cell_flag(t) = 2
            ghost_value(t) = sum_bvalue/sum_bvol
            t_bval(c) = ghost_value(t)
            cell_ivalue(t) = sum_value/sum_vol ; t_val(c) = sum_value/sum_vol
          else
            cell_ivalue(t) = sum_value/sum_vol ; t_val(c) = sum_value/sum_vol
          end if
          t_vol(c) = sum_vol ; t_bvol(c) = sum_bvol
        end if
      end do
    end do

    deallocate(t_bval, t_vol, t_bvol)

    return

  end subroutine set_mgbound

  subroutine set_mgseep(cell_flag)
  !*****************************************************************************
  ! set_mgseep -- Set mg condition seepage
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(inout) :: cell_flag(:)
    ! -- local
    real(DP) :: sum_area
    real(DP), allocatable :: t_val(:), t_are(:)
    !---------------------------------------------------------------------------
    allocate(t_val(nsurf), t_are(nsurf))
    t_val(:) = DZERO ; t_are(:) = DZERO

    do n = 1, nlevel
      do s = s_end(n-1)+1, s_end(n)
        c = surf2cell(s)
        if (n == 1 .and. out_flag(c) == 0) then
          t = surf2cals(s)
          t_val(s) = float(cell_flag(t))
          t_are(s) = area(c,1)
        else if (n > 1 .and. out_flag(c) == 0) then
          t = surf2cals(s)
          nc = s - s_end(n-1)
          if (mod(nc,nxgrid(n)) == 0) then
            i = nxgrid(n)*2 - 1
          else
            i = mod(nc,nxgrid(n))*2 - 1
          end if
          k = (nc-1)/(nxgrid(n)*nygrid(n))*2 + 1
          j = (nc-1)/nxgrid(n)*2 + 1 - (k-1)*nygrid(n)
          x1 = s_end(n-2) + nxgrid(n-1)*(j-1) + i
          x2 = x1 + 1 ; y1 = x1 + nxgrid(n-1) ; y2 = y1 + 1
          sum_value = t_val(x1)*t_are(x1) + t_val(x2)*t_are(x2) +&
                      t_val(y1)*t_are(y1) + t_val(y2)*t_are(y2)
          sum_area = t_are(x1) + t_are(x2) + t_are(y1) + t_are(y2)
          t_val(s) = sum_value/sum_area ; t_are(s) = sum_area
          if (sum_value > sum_area*DHALF) then
            cell_flag(t) = 1
          end if
        end if
      end do
    end do

    deallocate(t_val, t_are)

    return

  end subroutine set_mgseep

  subroutine set_resinfo(bflag)
  !*****************************************************************************
  ! set_resinfo -- Set restriction information
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: bflag(ncalc)
    ! -- local
    integer(I4) :: maxnum_restr, tfine_num, cc, nc_num, nf
    !---------------------------------------------------------------------------
    maxnum_restr = (calc_end(nlevel) - calc_end(1))*8
    allocate(restr_index(0:calc_end(nlevel)), restr_num(maxnum_restr))
    restr_index(:) = 0 ; restr_num(:) = 0
    allocate(restr_wei(maxnum_restr))
    restr_wei(:) = DZERO

    tfine_num = 0

    do n = 1, nlevel-1
      do nc = 1, c_end(n+1) - c_end(n)
        nc_num = c_end(n) + nc
        if (out_flag(nc_num) /= 1) then
          cc = cell2calc(nc_num)
          if (bflag(cc) /= 1) then
            if (mod(nc,nxgrid(n+1)) == 0) then
              fx = nxgrid(n+1)*2 - 1
            else
              fx = mod(nc,nxgrid(n+1))*2 - 1
            end if
            fz = (nc-1)/(nxgrid(n+1)*nygrid(n+1))*2 + 1
            fy = (nc-1)/nxgrid(n+1)*2 + 1 - (fz-1)*nygrid(n+1)

            do k = fz, fz+1
              do j = fy, fy+1
                do i = fx, fx+1
                  nf = c_end(n-1) + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-  1) + i
                  if (out_flag(nf) /= 1) then
                    fcal_num = cell2calc(nf)
                    if (bflag(fcal_num) /= 1) then
                      tfine_num = tfine_num + 1
                      restr_index(cc) = tfine_num
                      restr_num(tfine_num) = fcal_num
                      restr_wei(tfine_num) = vol_c(nf)
                    end if
                  end if
                end do
              end do
            end do
          end if
        end if
      end do
    end do

    restr_index(0) = 0

    return

  end subroutine set_resinfo

  subroutine set_proinfo(bflag)
  !*****************************************************************************
  ! set_proinfo -- Set prolongation information
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: bflag(ncalc)
    ! -- local
    integer(I4) :: dir
    integer(I4), allocatable :: f_num(:)
    !---------------------------------------------------------------------------
    allocate(coa_conn(FACE), f_num(8))
    allocate(prolo_index(0:calc_end(nlevel-1)), fine2coas(calc_end(nlevel-1)))
    allocate(prolo_num(calc_end(nlevel-1)*7))
    allocate(prolo_wei(calc_end(nlevel-1)*7))
    coa_conn(:) = 0 ; f_num(:) = 0 ; prolo_index(:) = 0
    fine2coas(:) = 0 ; prolo_num(:) = 0 ; prolo_wei(:) = DZERO

    tcoase_num = 0

    do n = 2, nlevel
      do i = calc_end(n-1)+1, calc_end(n)
        if (bflag(i) /= 1) then
          coa_conn(:) = 0
          do k = off_index(i-1)+1, off_index(i)
            dir = conn_dir(k) ; j = off_row(k) ; coa_conn(dir) = j
          end do
          c_num = calc2cell(i)
          coa_num = c_num - c_end(n-1)
          if (mod(coa_num,nxgrid(n)) == 0) then
            fx = nxgrid(n)*2 - 1
          else
            fx = mod(coa_num,nxgrid(n))*2 - 1
          end if
          fz = (coa_num-1)/(nxgrid(n)*nygrid(n))*2 + 1
          fy = (coa_num-1)/nxgrid(n)*2 + 1 - (fz-1)*nygrid(n)
          f_num(1) = c_end(n-2) + nxgrid(n-1)*nygrid(n-1)*(fz-1) + nxgrid(n-1)*(fy-1) + fx
          f_num(2) = f_num(1) + 1 ; f_num(3) = f_num(1) + nxgrid(n-1)
          f_num(4) = f_num(3) + 1 ; f_num(5) = f_num(1) + nxgrid(n-1)*nygrid(n-1)
          f_num(6) = f_num(5) + 1 ; f_num(7) = f_num(5) + nxgrid(n-1)
          f_num(8) = f_num(7) + 1
          do nc = 1, 8
            fcal_num = cell2calc(f_num(nc))
            if (fcal_num /= 0) then
              fine2coas(fcal_num) = i
              ! -- Set coefficent one (coef1)
                call set_coef1()
              ! -- Set coefficent three on diagonal (coef4dia)
                call set_coef3dia()
              ! -- Set coefficent nine on adjacent (coef4adj)
                call set_coef9adj()
              prolo_index(fcal_num) = tcoase_num
            end if
          end do
        end if
      end do
    end do

    prolo_index(0) = 0

    return

  end subroutine set_proinfo

  subroutine set_coef1()
  !*****************************************************************************
  ! set_coef1 -- Set coefficent one
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (coa_conn(1) > 0 .and. coa_conn(2) > 0 .and. coa_conn(3) > 0) then
      coa_num = c_num - nxgrid(n) - (nxgrid(n)*nygrid(n))  - 1
      if (nc == 1 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(2) > 0 .and. coa_conn(4) > 0) then
      coa_num = c_num - nxgrid(n) - (nxgrid(n)*nygrid(n))  + 1
      if (nc == 2 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(3) > 0 .and. coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n) - (nxgrid(n)*nygrid(n))  - 1
      if (nc == 3 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(4) > 0 .and. coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n) - (nxgrid(n)*nygrid(n))  + 1
      if (nc == 4 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(2) > 0 .and. coa_conn(3) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num - nxgrid(n) + (nxgrid(n)*nygrid(n))  - 1
      if (nc == 5 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(2) > 0 .and. coa_conn(4) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num - nxgrid(n) + (nxgrid(n)*nygrid(n))  + 1
      if (nc == 6 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(3) > 0 .and. coa_conn(5) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num + nxgrid(n) + (nxgrid(n)*nygrid(n))  - 1
      if (nc == 7 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    else if (coa_conn(4) > 0 .and. coa_conn(5) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num + nxgrid(n) + (nxgrid(n)*nygrid(n))  + 1
      if (nc == 8 .and. out_flag(coa_num) == 0) then
        tcoase_num = tcoase_num + 1
        prolo_num(tcoase_num) = cell2calc(coa_num)
        prolo_wei(tcoase_num) = DONE
      end if
    end if

    return

  end subroutine set_coef1

  subroutine set_coef3dia()
  !*****************************************************************************
  ! set_coef3dia -- Set coefficent three on diagonal
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (coa_conn(1) > 0 .and. coa_conn(2) > 0) then
      coa_num = c_num - nxgrid(n) - (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 2) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(3) > 0) then
      coa_num = c_num - 1 - (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 3) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(4) > 0) then
      coa_num = c_num + 1 - (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 2 .or. nc == 4) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(1) > 0 .and. coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n) - (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 3 .or. nc == 4) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(2) > 0 .and. coa_conn(3) > 0) then
      coa_num = c_num - nxgrid(n) - 1
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 5) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(2) > 0 .and. coa_conn(4) > 0) then
      coa_num = c_num - nxgrid(n) + 1
      if (out_flag(coa_num) == 0) then
        if (nc == 2 .or. nc == 6) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(2) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num - nxgrid(n) + (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 5 .or. nc == 6) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(3) > 0 .and. coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n) - 1
      if (out_flag(coa_num) == 0) then
        if (nc == 3 .or. nc == 7) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(3) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num - 1 + (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 5 .or. nc == 7) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(4) > 0 .and. coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n) + 1
      if (out_flag(coa_num) == 0) then
        if (nc == 4 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(4) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num + 1 + (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 6 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    else if (coa_conn(5) > 0 .and. coa_conn(6) > 0) then
      coa_num = c_num + nxgrid(n) + (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 7 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = cell2calc(coa_num)
          prolo_wei(tcoase_num) = 3.00_DP
        end if
      end if
    end if

    return

  end subroutine set_coef3dia

  subroutine set_coef9adj()
  !*****************************************************************************
  ! set_coef9adj -- Set coefficent nine on adjacent
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (coa_conn(1) > 0) then
      coa_num = c_num - (nxgrid(n)*nygrid(n))
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 2 .or. nc == 3 .or. nc == 4) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(1)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    else if (coa_conn(2) > 0) then
      coa_num = c_num - nxgrid(n)
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 2 .or. nc == 5 .or. nc == 6) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(2)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    else if (coa_conn(3) > 0) then
      coa_num = c_num - 1
      if (out_flag(coa_num) == 0) then
        if (nc == 1 .or. nc == 3 .or. nc == 5 .or. nc == 7) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(3)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    else if (coa_conn(4) > 0) then
      coa_num = c_num + 1
      if (out_flag(coa_num) == 0) then
        if (nc == 2 .or. nc == 4 .or. nc == 6 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(4)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    else if (coa_conn(5) > 0) then
      coa_num = c_num + nxgrid(n)
      if (out_flag(coa_num) == 0) then
        if (nc == 3 .or. nc == 4 .or. nc == 7 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(5)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    else if (coa_conn(6) > 0) then
      if (out_flag(coa_num) == 0) then
        coa_num = c_num + (nxgrid(n)*nygrid(n))
        if (nc == 5 .or. nc == 6 .or. nc == 7 .or. nc == 8) then
          tcoase_num = tcoase_num + 1
          prolo_num(tcoase_num) = coa_conn(6)
          prolo_wei(tcoase_num) = 9.00_DP
        end if
      end if
    end if

    return

  end subroutine set_coef9adj

end module set_mg_condition
