module write_outputfile
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: VARIABLELENGTH, FORMATLENGTH, OUTFORM, DZERO,&
                               MASSCHARA, FACE, DNOVAL, DHALF
  use read_inputfile, only: outhead_flag, outsrati_flag, outwlevel_flag,&
                            outmass_flag, outvelc_flag, wlout_flag, msout_flag,&
                            nxyz, calc_unit, outwell_flag, outrech_flag
  use set_classification, only: cell2calc, calc_end, calc2cell,&
                                out_flag, get_calgrd, cals_end
  use make_cell, only: dis
  use set_condition, only: off_index, off_row, conn_dir
  use set_calculation, only: bound_cflag, surf_head
  use allocate_module, only: head_new, hydcon_new, srati_new
  use time_module, only: timestep_num, now_time
  use check_simulation, only: check_outtiming, write_flag, lasttime_flag
  use make_output, only: wl_header, ms_header, wlout_tnum, msout_tnum, we_dnum,&
                         wlevel, con_sum, sto_sum, rec_sum, wel_sum, all_sum,&
                         vel_point, inp_we, act_we, dif_we, well_diflag,&
                         inp_re, act_re, dif_re, rech_diflag, re_dnum,&
                         set_pointwlhead, set_classmshead, set_pointmshead,&
                         set_wholemshead, allocate_wlevel, allocate_mass,&
                         allocate_velc, allocate_welldif, allocate_rechdif,&
                         calc_pointwl, calc_cellmass, calc_outmass,&
                         calc_outvelc, calc_welldif, calc_rechdif

  implicit none
  private
  public :: write_outf

  ! -- local
  integer(I4) :: header_flag = 0, allocate_flag = 0
  character(len=FORMATLENGTH) :: wlformat, msformat
  integer(I4), allocatable :: wi(:), wj(:), wk(:), ri(:), rj(:)

  integer(I4) :: i, j, dummy

  contains

  subroutine write_outf()
  !***************************************************************************
  ! write_outf -- write output file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Check output timing (outtiming)
      call check_outtiming()

    if (allocate_flag == 0) then
      allocate_flag = 1
      ! -- Allocate output variable (outvar)
        call allocate_outvar()
    end if

    if (header_flag == 0 .and. write_flag == 1) then
      header_flag = 1
      ! -- Write output header (header)
        call write_header()
    end if

    if (outmass_flag > 0) then
      ! -- Calculate cell massbalance (cellmass)
        call calc_cellmass()
    end if

    if (outwell_flag > 0) then
      ! -- Calculate well difference (welldif)
        call calc_welldif()
    end if

    if (outrech_flag > 0) then
      ! -- Calculate recharge difference (rechdif)
        call calc_rechdif()
    end if

    if (write_flag == 1) then
      if (outhead_flag > 0 .and. mod(timestep_num,outhead_flag) == 0) then
        ! -- Write head file (outheadf)
          call write_outheadf()
      else if (outhead_flag > 0 .and. lasttime_flag == 1) then
        ! -- Write head file (outheadf)
          call write_outheadf()
      end if

      if (outsrati_flag > 0 .and. mod(timestep_num,outsrati_flag) == 0) then
        ! -- Write saturation file (outsratif)
          call write_outsratif()
      else if (outsrati_flag > 0 .and. lasttime_flag == 1) then
        ! -- Write saturation file (outsratif)
          call write_outsratif()
      end if

      if (outwlevel_flag > 0 .and. mod(timestep_num,outwlevel_flag) == 0) then
        ! -- Write waterlevel file (outwlevelf)
          call write_outwlevelf()
      else if (outwlevel_flag > 0 .and. lasttime_flag== 1) then
        ! -- Write waterlevel file (outwlevelf)
          call write_outwlevelf()
      end if

      if (outmass_flag > 0 .and. mod(timestep_num,outmass_flag) == 0) then
        ! -- Write massbalance file (outmassf)
          call write_outmassf()
      else if (outmass_flag > 0 .and. lasttime_flag== 1) then
        ! -- Write massbalance file (outmassf)
          call write_outmassf()
      end if

      if (outvelc_flag > 0 .and. mod(timestep_num,outvelc_flag) == 0) then
        ! -- Write velocity file (outvelcf)
          call write_outvelcf()
      else if (outvelc_flag > 0 .and. lasttime_flag== 1) then
        ! -- Write velocity file (outvelcf)
          call write_outvelcf()
      end if

      if (outwell_flag > 0 .and. mod(timestep_num,outwell_flag) == 0) then
        ! -- Write well pumping results file (outwellf)
          call write_outwellf()
      else if (outwell_flag > 0 .and. lasttime_flag== 1) then
        ! -- Write well pumping results file (outwellf)
          call write_outwellf()
      end if

      if (outrech_flag > 0 .and. mod(timestep_num,outrech_flag) == 0) then
        ! -- Write recharge results file (outrechf)
          call write_outrechf()
      else if (outrech_flag > 0 .and. lasttime_flag== 1) then
        ! -- Write recharge results file (outrechf)
          call write_outrechf()
      end if

    end if

    return

  end subroutine write_outf

  subroutine allocate_outvar()
  !***************************************************************************
  ! allocate_outvar -- Allocate output variable (outvar)
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! waterlevel file
    if (outwlevel_flag > 0) then
      if (wlout_flag >= 1) then
        ! -- Set point water level header (pointwl)
          call set_pointwlhead()
      end if

      ! -- Allocate water level (wlevel)
        call allocate_wlevel
    end if

    ! massbalance file
    if (outmass_flag > 0) then
      if (msout_flag == 0) then
        ! -- Set classification massbalance header (classmshead)
          call set_classmshead()
      else if (msout_flag == 1) then
        ! -- Set point massbalance header (pointmshead)
          call set_pointmshead()
      else if (msout_flag == 2) then
        ! -- Set whole area massbalance header (wholemshead)
          call set_wholemshead()
      end if

      ! -- Allocate massbalance (mass)
        call allocate_mass
    end if

    ! velocity file
    if (outvelc_flag > 0) then
      ! -- Allocate velocity level (velc)
        call allocate_velc
    end if

    ! well pumping results file
    if (outwell_flag > 0) then
      ! -- Allocate well difference (welldif)
        call allocate_welldif
      allocate(wi(calc_end(1)), wj(calc_end(1)), wk(calc_end(1)))
      wi(:) = 0 ; wj(:) = 0 ; wk(:) = 0
      do i = 1, calc_end(1)
        ! -- Get relationship of calculation & grid number (calgrd)
          call get_calgrd(i, wi(i), wj(i), wk(i))
      end do
    end if

    ! rech results file
    if (outrech_flag > 0) then
      ! -- Allocate recharge difference (rechdif)
        call allocate_rechdif
      allocate(ri(cals_end(1)), rj(cals_end(1)))
      ri(:) = 0 ; rj(:) = 0
      do i = 1, cals_end(1)
        ! -- Get relationship of calculation & grid number (calgrd)
          call get_calgrd(i, ri(i), rj(i), dummy)
      end do
    end if

    return

  end subroutine allocate_outvar

  subroutine write_header()
  !***************************************************************************
  ! write_header -- Write output header
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    character(len=VARIABLELENGTH) :: str_wltnum, str_mstnum
    integer(I4) :: msout_tnum_format
    !---------------------------------------------------------------------------
    ! waterlevel file
    if (outwlevel_flag > 0) then
      write(105,'(3a)') trim(adjustl(calc_unit)), ",", trim(adjustl(wl_header))

      write(str_wltnum, *) wlout_tnum
      wlformat = "(" // OUTFORM // "," // trim(adjustl(str_wltnum)) // "(','," // OUTFORM // "))"
    end if

    ! massbalance file
    if (outmass_flag > 0) then
      write(106,'(3a)') trim(adjustl(calc_unit)), ",", trim(adjustl(ms_header))

      write(str_mstnum, *) msout_tnum
      do i = 1, msout_tnum
        if (i == 1) then
          ms_header = "," // trim(adjustl(MASSCHARA))
        else
          ms_header = trim(adjustl(ms_header)) // "," //  trim(adjustl(MASSCHARA))
        end if
      end do

      write(106,'(a)') (trim(adjustl(ms_header)))

      msout_tnum_format = msout_tnum*5

      write(str_mstnum, *) msout_tnum_format
      msformat = "(" // OUTFORM // "," // trim(adjustl(str_mstnum)) // "(','," // OUTFORM // "))"

    end if

    ! well pumping results file
    if (outwell_flag > 0) then
      write(108,'(2a)') trim(adjustl(calc_unit)), "   The number of difference"
      write(108,'(a)') "    I    J    K  INPUT VALUE ACTUAL VALUE   DIFFERENCE"
    end if

    ! recharge results file
    if (outrech_flag > 0) then
      write(109,'(2a)') trim(adjustl(calc_unit)), "   The number of difference"
      write(109,'(a)') "    I    J  INPUT VALUE ACTUAL VALUE   DIFFERENCE"
    end if

    return

  end subroutine write_header

  subroutine write_outheadf()
  !***************************************************************************
  ! write_outheadf -- Write head file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    write(102) now_time, (head_new(i), i = 1, calc_end(1))

    return

  end subroutine write_outheadf

  subroutine write_outsratif()
  !***************************************************************************
  ! write_outsratif -- Write saturation file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    write(103) now_time, (srati_new(i), i = 1, calc_end(1))

    return

  end subroutine write_outsratif

  subroutine write_outwlevelf()
  !***************************************************************************
  ! write_outwlevelf -- Write waterlevel file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (wlout_flag >= 1) then
      ! -- Calculate point water level (pointwl)
        call calc_pointwl()
    end if

    write(105,trim(adjustl(wlformat))) now_time, (wlevel(i), i = 1, wlout_tnum)

    return

  end subroutine write_outwlevelf

  subroutine write_outmassf()
  !***************************************************************************
  ! write_outmassf -- Write massbalance file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calculate output massbalance (outmass)
      call calc_outmass()

    write(106,trim(adjustl(msformat))) now_time, (con_sum(i), sto_sum(i),&
          rec_sum(i), wel_sum(i), all_sum(i), i = 1, msout_tnum)

    return

  end subroutine write_outmassf

  subroutine write_outvelcf()
  !***************************************************************************
  ! write_outvelcf -- Write velocity file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calculate output velocity (outvelc)
      call calc_outvelc()

    write(107) now_time, ((vel_point(i,j), i = 1, calc_end(1)), j = 1, 3)

    return

  end subroutine write_outvelcf

  subroutine write_outwellf()
  !***************************************************************************
  ! write_outwellf -- Write well pumping results file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    ! -- format
    81 format(es12.6,i10)
    82 format(3(i5),3(es13.4))
    !---------------------------------------------------------------------------
    if (we_dnum /= 0) then
      write(108,81) now_time, we_dnum
    end if

    do i = 1, calc_end(1)
      if (well_diflag(i) == 1) then
        write(108,82) wi(i), wj(i), wk(i), inp_we(i), act_we(i), dif_we(i)
      end if
    end do

    we_dnum = 0 ; well_diflag(:) = 0
    inp_we(:) = DZERO ; act_we(:) = DZERO ; dif_we(:) = DZERO

    return

  end subroutine write_outwellf

  subroutine write_outrechf()
  !***************************************************************************
  ! write_outrechf -- Write recharge results file
  !***************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    ! -- format
    91 format(es12.6,i10)
    92 format(2(i5),3(es13.4))
    !---------------------------------------------------------------------------
    if (re_dnum /= 0) then
      write(109,91) now_time, re_dnum
    end if

    do i = 1, cals_end(1)
      if (rech_diflag(i) == 1) then
        write(109,92) ri(i), rj(i), inp_re(i), act_re(i), dif_re(i)
      end if
    end do

    re_dnum = 0 ; rech_diflag(:) = 0
    inp_re(:) = DZERO ; act_re(:) = DZERO ; dif_re(:) = DZERO

    return

  end subroutine write_outrechf

end module write_outputfile
